"""A dataset class used to train Parametric UMAP models."""

from __future__ import annotations

import pathlib
import warnings

import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.sparse._coo
import torchvision.transforms
from pynndescent import NNDescent
from sklearn.utils import check_random_state
from torch.utils.data import Dataset

from ... import transforms as vak_transforms

# isort: off
# Ignore warnings from Numba deprecation:
# https://numba.readthedocs.io/en/stable/reference/deprecation.html#deprecation-of-object-mode-fall-back-behaviour-when-using-jit
# Numba is required by UMAP.
from numba.core.errors import NumbaDeprecationWarning

warnings.simplefilter("ignore", category=NumbaDeprecationWarning)
from umap.umap_ import fuzzy_simplicial_set  # noqa: E402

# isort: on


def get_umap_graph(
    X: npt.NDArray,
    n_neighbors: int = 10,
    metric: str = "euclidean",
    random_state: np.random.RandomState | None = None,
    max_candidates: int = 60,
    verbose: bool = True,
) -> scipy.sparse._coo.coo_matrix:
    r"""Get graph used by UMAP,
    the fuzzy topological representation.

    Parameters
    ----------
    X : numpy.ndarray
        Data from which to build the graph.
    n_neighbors : int
        Number of nearest neighbors to use
        when computing approximate nearest neighbors.
        Parameter passed to :class:`pynndescent.NNDescent`
        and :func:`umap._umap.fuzzy_simplicial_set`.
    metric : str
        Distance metric. Default is "cosine".
        Parameter passed to :class:`pynndescent.NNDescent`
        and :func:`umap._umap.fuzzy_simplicial_set`.
    random_state : numpy.random.RandomState
        Either a numpy.random.RandomState instance,
        or None.
    max_candidates : int
        Default is 60.
        Parameter passed to :class:`pynndescent.NNDescent`.
    verbose : bool
        Whether :class:`pynndescent.NNDescent` should log
        finding the approximate nearest neighbors.
        Default is True.

    Returns
    -------
    graph : scipy.sparse.csr_matrix

    Notes
    -----
    Adapted from https://github.com/timsainb/ParametricUMAP_paper

    The graph returned is a graph of the probabilities of an edge exists between points.

    Local, one-directional, probabilities (:math:`P^{UMAP}_{i|j}`)
    are computed between a point and its neighbors to determine
    the probability with which an edge (or simplex) exists,
    based upon an assumption that data is uniformly distributed
    across a manifold in a warped dataspace.
    Under this assumption, a local notion of distance
    is set by the distance to the :math:`k^{th}` nearest neighbor
    and the local probability is scaled by that local notion of distance.

    Where :math:`\rho_{i}` is a local connectivity parameter set
    to the distance from :math:`x_i` to its nearest neighbor,
    and :math:`\sigma_{i}` is a local connectivity parameter
    set to match the local distance around :math:`x_i` upon its :math:`k` nearest neighbors
    (where :math:`k` is a hyperparameter).
    In the UMAP package, these are calculated using :func:`umap._umap.smooth_knn_dist`.
    """
    random_state = (
        check_random_state(None) if random_state is None else random_state
    )

    # number of trees in random projection forest
    n_trees = 5 + int(round((X.shape[0]) ** 0.5 / 20.0))

    # max number of nearest neighbor iters to perform
    n_iters = max(5, int(round(np.log2(X.shape[0]))))

    # get nearest neighbors
    nnd = NNDescent(
        X.reshape((len(X), np.prod(np.shape(X)[1:]))),
        n_neighbors=n_neighbors,
        metric=metric,
        n_trees=n_trees,
        n_iters=n_iters,
        max_candidates=max_candidates,
        verbose=verbose,
    )

    # get indices and distances for 10 nearest neighbors of every point in dataset
    knn_indices, knn_dists = nnd.neighbor_graph

    # build fuzzy simplicial complex
    umap_graph, sigmas, rhos = fuzzy_simplicial_set(
        X=X,
        n_neighbors=n_neighbors,
        metric=metric,
        random_state=random_state,
        knn_indices=knn_indices,
        knn_dists=knn_dists,
    )

    return umap_graph


def get_graph_elements(
    graph: scipy.sparse._coo.coo_matrix, n_epochs: int
) -> tuple[
    scipy.sparse._coo.coo_matrix,
    npt.NDArray,
    npt.NDArray,
    npt.NDArray,
    npt.NDArray,
    int,
]:
    """Get graph elements for Parametric UMAP Dataset.

    Parameters
    ----------
    graph : scipy.sparse.csr_matrix
        The graph returned by :func:`get_umap_graph`.
    n_epochs : int
        Number of epochs model will be trained

    Returns
    -------
    graph : scipy.sparse._coo.coo_matrix
        The graph, now in COOrdinate format.
    epochs_per_sample : int
    head : numpy.ndarray
        Graph rows.
    tail : numpy.ndarray
        Graph columns.
    weight : numpy.ndarray
        Graph data.
    n_vertices : int
        Number of vertices in dataset.
    """
    graph = graph.tocoo()

    # eliminate duplicate entries by summing them together
    graph.sum_duplicates()

    # number of vertices in dataset
    n_vertices = graph.shape[1]

    # get the number of epochs based on the size of the dataset
    if n_epochs is None:
        # For smaller datasets we can use more epochs
        if graph.shape[0] <= 10000:
            n_epochs = 500
        else:
            n_epochs = 200

    # remove elements with very low probability
    graph.data[graph.data < (graph.data.max() / float(n_epochs))] = 0.0
    graph.eliminate_zeros()

    # get epochs per sample based upon edge probability
    epochs_per_sample = n_epochs * graph.data

    head = graph.row
    tail = graph.col
    weight = graph.data

    return graph, epochs_per_sample, head, tail, weight, n_vertices


class Datapipe(Dataset):
    """A datapipe used with Parametric UMAP models."""

    def __init__(
        self,
        dataset_path: str | pathlib.Path,
        dataset_df: pd.DataFrame,
        split: str,
        subset: str | None = None,
        n_epochs: int = 200,
        n_neighbors: int = 10,
        metric: str = "euclidean",
        random_state: int | None = None,
    ):
        """Initialize a :class:`ParametricUMAPDataset` instance.

        Parameters
        ----------
        dataset_path : pathlib.Path
            Path to directory that represents a
            parametric UMAP dataset,
            as created by
            :func:`vak.prep.prep_parametric_umap_dataset`.
        dataset_df : pandas.DataFrame
            A parametric UMAP dataset,
            represented as a :class:`pandas.DataFrame`.
        split : str
            The name of a split from the dataset,
            one of {'train', 'val', 'test'}.
        subset : str, optional
            Name of subset to use.
            If specified, this takes precedence over split.
            Subsets are typically taken from the training data
            for use when generating a learning curve.
        n_epochs : int
            Number of epochs model will be trained. Default is 200.
        transform : callable, optional
        """
        # subset takes precedence over split, if specified
        if subset:
            dataset_df = dataset_df[dataset_df.subset == subset].copy()
        else:
            dataset_df = dataset_df[dataset_df.split == split].copy()

        data = np.stack(
            [
                np.load(dataset_path / spect_path)
                for spect_path in dataset_df.spect_path.values
            ]
        )

        graph = get_umap_graph(
            data,
            n_neighbors=n_neighbors,
            metric=metric,
            random_state=random_state,
        )

        (
            graph,
            epochs_per_sample,
            head,
            tail,
            _,
            _,
        ) = get_graph_elements(graph, n_epochs)

        # we repeat each sample in (head, tail) a certain number of times depending on its probability
        self.edges_to_exp, self.edges_from_exp = (
            np.repeat(head, epochs_per_sample.astype("int")),
            np.repeat(tail, epochs_per_sample.astype("int")),
        )
        # we then shuffle -- not sure this is necessary if the dataset is shuffled during training?
        shuffle_mask = np.random.permutation(np.arange(len(self.edges_to_exp)))
        self.edges_to_exp = self.edges_to_exp[shuffle_mask].astype(np.int64)
        self.edges_from_exp = self.edges_from_exp[shuffle_mask].astype(
            np.int64
        )

        self.data = data
        self.dataset_df = dataset_df
        self.transform = torchvision.transforms.Compose(
            [
                vak_transforms.ToFloatTensor(),
                vak_transforms.AddChannel(),
            ]
        )

    @property
    def duration(self):
        return self.dataset_df["duration"].sum()

    def __len__(self):
        return self.edges_to_exp.shape[0]

    @property
    def shape(self):
        tmp_x_ind = 0
        tmp_item = self.__getitem__(tmp_x_ind)
        return tmp_item[0].shape

    def __getitem__(self, index):
        edges_to_exp = self.data[self.edges_to_exp[index]]
        edges_from_exp = self.data[self.edges_from_exp[index]]
        edges_to_exp = self.transform(edges_to_exp)
        edges_from_exp = self.transform(edges_from_exp)
        return (edges_to_exp, edges_from_exp)

    @classmethod
    def from_dataset_path(
        cls,
        dataset_path: str | pathlib.Path,
        split: str,
        subset: str | None = None,
        n_neighbors: int = 10,
        metric: str = "euclidean",
        random_state: int | None = None,
        n_epochs: int = 200,
    ):
        """Make a :class:`ParametricUMAPDataset` instance,
        given the path to parametric UMAP dataset.

        Parameters
        ----------
        dataset_path : pathlib.Path
            Path to directory that represents a
            parametric UMAP dataset,
            as created by
            :func:`vak.prep.prep_parametric_umap_dataset`.
        split : str
            The name of a split from the dataset,
            one of {'train', 'val', 'test'}.
        subset : str, optional
            Name of subset to use.
            If specified, this takes precedence over split.
            Subsets are typically taken from the training data
            for use when generating a learning curve.
        n_neighbors : int
            Number of nearest neighbors to use
            when computing approximate nearest neighbors.
            Parameter passed to :class:`pynndescent.NNDescent`
            and :func:`umap._umap.fuzzy_simplicial_set`.
        metric : str
            Distance metric. Default is "cosine".
            Parameter passed to :class:`pynndescent.NNDescent`
            and :func:`umap._umap.fuzzy_simplicial_set`.
        random_state : numpy.random.RandomState
            Either a numpy.random.RandomState instance,
            or None.

        Returns
        -------
        dataset : vak.datasets.parametric_umap.TrainDatapipe
        """
        import vak.datapipes  # import here just to make classmethod more explicit

        dataset_path = pathlib.Path(dataset_path)
        metadata = vak.datapipes.parametric_umap.Metadata.from_dataset_path(
            dataset_path
        )

        dataset_csv_path = dataset_path / metadata.dataset_csv_filename
        dataset_df = pd.read_csv(dataset_csv_path)

        return cls(
            dataset_path,
            dataset_df,
            split,
            subset,
            n_epochs,
            n_neighbors,
            metric,
            random_state,
        )



"""Class that represent a model builit into ``vak``."""

from __future__ import annotations

import inspect
from typing import Callable, Type

import lightning
import torch

from .decorator import ModelDefinitionValidationError
from .definition import validate as validate_definition


class ModelFactory:
    """Class that represent a model builit into ``vak``.

    Attributes
    ----------
    definition: vak.models.definition.ModelDefinition
    family: lighting.pytorch.LightningModule

    Notes
    -----
    This class is used by the :func:`vak.models.decorator.model`
    decorator to make a new class representing a model
    from a model definition.
    As such, this class is not meant to be used directly.
    See the docstring of :func:`vak.models.decorator.model`
    for more detail.
    """

    def __init__(
        self,
        definition: Type,
        family: lightning.pytorch.LightningModule,
    ) -> None:
        if not issubclass(family, lightning.pytorch.LightningModule):
            raise TypeError(
                "The ``family`` argument to the ``vak.models.model`` decorator"
                "should be a subclass of ``lightning.pytorch.LightningModule``,"
                f"but the type was: {type(family)}, "
                "which was not recognized as a subclass "
                "of ``lightning.pytorch.LightningModule``."
            )

        try:
            validate_definition(definition)
        except ValueError as err:
            raise ModelDefinitionValidationError(
                f"Validation failed for the following model definition:\n{definition}"
            ) from err
        except TypeError as err:
            raise ModelDefinitionValidationError(
                f"Validation failed for the following model definition:\n{definition}"
            ) from err

        self.definition = definition
        self.family = family

    def attributes_from_config(self, config: dict):
        """Get attributes for an instance of a model,
        given a configuration.

        Given a :class:`dict`, ``config``, return instances
        of `network`, `optimizer`, `loss`, and `metrics`.

        Parameters
        ----------
        config : dict
            A :class:`dict` obtained by calling
            :meth:`vak.config.ModelConfig.to_dict()`.

        Returns
        -------
        network : torch.nn.Module, dict
            An instance of a ``torch.nn.Module``
            that implements a neural network,
            or a ``dict`` that maps human-readable string names
            to a set of such instances.
        loss : torch.nn.Module, callable
            An instance of a ``torch.nn.Module``
            that implements a loss function.
        optimizer : torch.optim.Optimizer
            An instance of a ``torch.optim.Optimizer`` class
            used with ``loss`` to optimize
            the parameters of ``network``.
        metrics : dict
            A ``dict`` that maps human-readable string names
            to ``Callable`` functions, used to measure
            performance of the model.
        """
        network_kwargs = config.get(
            "network", self.definition.default_config["network"]
        )
        if inspect.isclass(self.definition.network):
            network = self.definition.network(**network_kwargs)
        elif isinstance(self.definition.network, dict):
            network = {}
            for net_name, net_class in self.definition.network.items():
                net_class_kwargs = network_kwargs.get(net_name, {})
                network[net_name] = net_class(**net_class_kwargs)

        if isinstance(self.definition.network, dict):
            params = [
                param
                for net_name, net_instance in network.items()
                for param in net_instance.parameters()
            ]
        else:
            params = network.parameters()

        optimizer_kwargs = config.get(
            "optimizer", self.definition.default_config["optimizer"]
        )
        optimizer = self.definition.optimizer(
            params=params, **optimizer_kwargs
        )

        if inspect.isclass(self.definition.loss):
            loss_kwargs = config.get(
                "loss", self.definition.default_config["loss"]
            )
            loss = self.definition.loss(**loss_kwargs)
        else:
            loss = self.definition.loss

        metrics_config = config.get(
            "metrics", self.definition.default_config["metrics"]
        )
        metrics = {}
        for metric_name, metric_class in self.definition.metrics.items():
            metrics_class_kwargs = metrics_config.get(metric_name, {})
            metrics[metric_name] = metric_class(**metrics_class_kwargs)

        return network, loss, optimizer, metrics

    def validate_init(
        self,
        network: torch.nn.Module | dict | None = None,
        loss: torch.nn.Module | Callable | None = None,
        optimizer: torch.optim.Optimizer | None = None,
        metrics: dict | None = None,
    ):
        """Validate arguments to ``vak.models.base.Model.init``.

        Parameters
        ----------
        network : torch.nn.Module, dict
            An instance of a ``torch.nn.Module``
            that implements a neural network,
            or a ``dict`` where each key is a string
             that maps a human-readable name
            to a ``torch.nn.Module`` instance.
        loss : torch.nn.Module, callable
            An instance of a ``torch.nn.Module``
            that implements a loss function,
            or a callable Python function that
            computes a scalar loss.
        optimizer : torch.optim.Optimizer
            An instance of a ``torch.optim.Optimizer`` class
            used with ``loss`` to optimize
            the parameters of ``network``.
        metrics : dict
            A ``dict`` that maps human-readable string names
            to ``Callable`` functions, used to measure
            performance of the model.

        Returns
        -------
        None

        This method does not return values;
        it just raises an error if any value is invalid.
        """
        if network:
            if inspect.isclass(self.definition.network):
                if not isinstance(network, self.definition.network):
                    raise TypeError(
                        f"``network`` should be an instance of {self.definition.network}"
                        f"but was of type {type(network)}"
                    )

            elif isinstance(self.definition.network, dict):
                if not isinstance(network, dict):
                    raise TypeError(
                        "Expected ``network`` to be a ``dict`` mapping network names "
                        f"to ``torch.nn.Module`` instances, but type was {type(network)}"
                    )
                expected_network_dict_keys = list(
                    self.definition.network.keys()
                )
                network_dict_keys = list(network.keys())
                if not all(
                    [
                        expected_network_dict_key in network_dict_keys
                        for expected_network_dict_key in expected_network_dict_keys
                    ]
                ):
                    missing_keys = set(expected_network_dict_keys) - set(
                        network_dict_keys
                    )
                    raise ValueError(
                        f"The following keys were missing from the ``network`` dict: {missing_keys}"
                    )
                if any(
                    [
                        network_dict_key not in expected_network_dict_keys
                        for network_dict_key in network_dict_keys
                    ]
                ):
                    extra_keys = set(network_dict_keys) - set(
                        expected_network_dict_keys
                    )
                    raise ValueError(
                        f"The following keys in the ``network`` dict are not valid: {extra_keys}."
                        f"Valid keys are: {expected_network_dict_keys}"
                    )

                for network_name, network_instance in network.items():
                    if not isinstance(
                        network_instance, self.definition.network[network_name]
                    ):
                        raise TypeError(
                            f"Network with name '{network_name}' in ``network`` dict "
                            f"should be an instance of {self.definition.network[network_name]}"
                            f"but was of type {type(network)}"
                        )
            else:
                raise TypeError(
                    f"Invalid type for ``network``: {type(network)}"
                )

        if loss:
            if issubclass(self.definition.loss, torch.nn.Module):
                if not isinstance(loss, self.definition.loss):
                    raise TypeError(
                        f"``loss`` should be an instance of {self.definition.loss}"
                        f"but was of type {type(loss)}"
                    )
            elif callable(self.definition.loss):
                if loss is not self.definition.loss:
                    raise ValueError(
                        f"``loss`` should be the following callable (probably a function): {self.definition.loss}"
                    )
            else:
                raise TypeError(f"Invalid type for ``loss``: {type(loss)}")

        if optimizer:
            if not isinstance(optimizer, self.definition.optimizer):
                raise TypeError(
                    f"``optimizer`` should be an instance of {self.definition.optimizer}"
                    f"but was of type {type(optimizer)}"
                )

        if metrics:
            if not isinstance(metrics, dict):
                raise TypeError(
                    "``metrics`` should be a ``dict`` mapping string metric names "
                    f"to callable metrics, but type of ``metrics`` was {type(metrics)}"
                )
            for metric_name, metric_callable in metrics.items():
                if metric_name not in self.definition.metrics:
                    raise ValueError(
                        f"``metrics`` has name '{metric_name}' but that name "
                        f"is not in the model definition. "
                        f"Valid metric names are: {', '.join(list(self.definition.metrics.keys()))}"
                    )

                if not isinstance(
                    metric_callable, self.definition.metrics[metric_name]
                ):
                    raise TypeError(
                        f"metric '{metric_name}' should be an instance of {self.definition.metrics[metric_name]}"
                        f"but was of type {type(metric_callable)}"
                    )

    def validate_instances_or_get_default(
        self,
        network: torch.nn.Module | dict | None = None,
        loss: torch.nn.Module | Callable | None = None,
        optimizer: torch.optim.Optimizer | None = None,
        metrics: dict | None = None,
    ):
        """Validate instances of model attributes, using its definition,
        or if no instance is passed in for an attribute,
        make an instance using the default config.

        Takes in instances of the attributes defined by the class variable
        ``self.definition``: ``network``, ``loss``, ``optimizer``, and ``metrics``.
        If any of those arguments are ``None``, then ``__init__``
        instantiates the corresponding attribute with its defaults.
        If any of those arguments are not an instance of the type
        defined by ``self.definition``, then a TypeError is raised.

        Parameters
        ----------
        network : torch.nn.Module, dict
            An instance of a ``torch.nn.Module``
            that implements a neural network,
            or a ``dict`` that maps human-readable string names
            to a set of such instances.
        loss : torch.nn.Module, callable
            An instance of a ``torch.nn.Module``
            that implements a loss function,
            or a callable Python function that
            computes a scalar loss.
        optimizer : torch.optim.Optimizer
            An instance of a ``torch.optim.Optimizer`` class
            used with ``loss`` to optimize
            the parameters of ``network``.
        metrics : dict
            A ``dict`` that maps human-readable string names
            to ``Callable`` functions, used to measure
            performance of the model.
        """
        from .decorator import ModelDefinitionValidationError

        # check that we are a sub-class of some other class with required class variables
        if not hasattr(self, "definition"):
            raise ValueError(
                "This model does not have a definition."
                "Define a model by wrapping a class with the required class variables with "
                "a ``vak.models`` decorator, e.g. ``vak.models.windowed_frame_classification_model``"
            )

        try:
            validate_definition(self.definition)
        except ModelDefinitionValidationError as err:
            raise ValueError(
                "Creating model instance failed because model definition is invalid."
            ) from err

        # ---- validate any instances that user passed in
        self.validate_init(network, loss, optimizer, metrics)

        if network is None:
            net_kwargs = self.definition.default_config.get("network")
            if isinstance(self.definition.network, dict):
                network = {
                    network_name: network_class(**net_kwargs[network_name])
                    for network_name, network_class in self.definition.network.items()
                }
            else:
                network = self.definition.network(**net_kwargs)

        if loss is None:
            if inspect.isclass(self.definition.loss):
                loss_kwargs = self.definition.default_config.get("loss")
                loss = self.definition.loss(**loss_kwargs)
            elif inspect.isfunction(self.definition.loss):
                loss = self.definition.loss

        if optimizer is None:
            optimizer_kwargs = self.definition.default_config.get("optimizer")
            if isinstance(network, dict):
                params = [
                    param
                    for net_name, net_instance in network.items()
                    for param in net_instance.parameters()
                ]
            else:
                params = network.parameters()
            optimizer = self.definition.optimizer(
                params=params, **optimizer_kwargs
            )

        if metrics is None:
            metric_kwargs = self.definition.default_config.get("metrics")
            metrics = {}
            for metric_name, metric_class in self.definition.metrics.items():
                metric_class_kwargs = metric_kwargs.get(metric_name, {})
                metrics[metric_name] = metric_class(**metric_class_kwargs)
        return network, loss, optimizer, metrics

    def from_config(self, config: dict, **kwargs):
        """Return a a new instance of a model, given a config :class:`dict`.

        Parameters
        ----------
        config : dict
            The dict obtained by by calling :meth:`vak.config.ModelConfig.asdict`.

        Returns
        -------
        model : lightning.LightningModule
            An instance of the model :attr:`~ModelFactory.family`
            with attributes specified by :attr:`~ModelFactory.definition`,
            that are initialized using parameters from ``config``.
        """
        network, loss, optimizer, metrics = self.attributes_from_config(config)
        network, loss, optimizer, metrics = (
            self.validate_instances_or_get_default(
                network,
                loss,
                optimizer,
                metrics,
            )
        )
        return self.family(
            network=network,
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            **kwargs,
        )

    def from_instances(
        self,
        network: torch.nn.Module | dict | None = None,
        loss: torch.nn.Module | Callable | None = None,
        optimizer: torch.optim.Optimizer | None = None,
        metrics: dict | None = None,
        **kwargs,
    ):
        """

        Parameters
        ----------
        network : torch.nn.Module, dict
            An instance of a ``torch.nn.Module``
            that implements a neural network,
            or a ``dict`` that maps human-readable string names
            to a set of such instances.
        loss : torch.nn.Module, callable
            An instance of a ``torch.nn.Module``
            that implements a loss function,
            or a callable Python function that
            computes a scalar loss.
        optimizer : torch.optim.Optimizer
            An instance of a ``torch.optim.Optimizer`` class
            used with ``loss`` to optimize
            the parameters of ``network``.
        metrics : dict
            A ``dict`` that maps human-readable string names
            to ``Callable`` functions, used to measure
            performance of the model.
        """
        network, loss, optimizer, metrics = (
            self.validate_instances_or_get_default(
                network,
                loss,
                optimizer,
                metrics,
            )
        )
        return self.family(
            network=network,
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            **kwargs,
        )



from . import (
    __main__,
    cli,
    common,
    config,
    datapipes,
    datasets,
    eval,
    learncurve,
    metrics,
    models,
    nets,
    nn,
    plot,
    predict,
    prep,
    train,
    transforms,
)
from .__about__ import (
    __author__,
    __commit__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
    __version__,
)

__all__ = [
    "__main__",
    "__author__",
    "__commit__",
    "__copyright__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__uri__",
    "__version__",
    "cli",
    "common",
    "config",
    "datapipes",
    "datasets",
    "eval",
    "learncurve",
    "metrics",
    "models",
    "nets",
    "nn",
    "plot",
    "predict",
    "prep",
    "train",
    "transforms",
]



"""Functionality to prepare splits of frame classification datasets
to generate a learning curve."""

from __future__ import annotations

import logging
import pathlib
from typing import Sequence

import attrs
import dask.bag as db
import numpy as np
import pandas as pd
from dask.diagnostics import ProgressBar

from ... import common, datapipes
from .. import split

logger = logging.getLogger(__name__)


@attrs.define(frozen=True)
class Sample:
    """Dataclass representing one sample
    in a frame classification dataset.

    Used to add paths for arrays from the sample
    to a ``dataset_df``, and to build
    the ``sample_ids`` vector and ``inds_in_sample`` vector
    for the entire dataset."""

    source_id: int = attrs.field()
    sample_id_vec: np.ndarray
    inds_in_sample_vec: np.ndarray


def make_index_vectors_for_each_subset(
    subsets_df: pd.DataFrame,
    dataset_path: str | pathlib.Path,
    input_type: str,
) -> pd.DataFrame:
    r"""Make npy files containing indexing vectors
    for each subset of the training data
    used to generate a learning curve
    with a frame classification dataset.

    This function is basically the same as
    :func:`vak.prep.frame_classification.make_splits.make_splits`,
    *except* that it only makes the indexing vectors
    for each subset of the training data.
    These indexing vectors are needed for each subset
    to properly grab windows from the npy files during training.
    There is no need to remake the npy files themselves though.

    All the indexing vectors for each split are saved
    in the "train" directory split inside ``dataset_path``.

    The indexing vectors are used by
    :class:`vak.datasets.frame_classification.TrainDatapipe`
    and :class:`vak.datasets.frame_classification.InferDatapipe`.
    These vectors make it possible to work with files,
    to avoid loading the entire dataset into memory,
    and to avoid working with memory-mapped arrays.
    The first is the ``sample_ids`` vector,
    that represents the "ID" of any sample :math:`(x, y)` in the split.
    We use these IDs to load the array files corresponding to the samples.
    For a split with :math:`m` samples, this will be an array of length :math:`T`,
    the total number of frames across all samples,
    with elements :math:`i \in (0, 1, ..., m - 1)`
    indicating which frames correspond to which sample :math:`m_i`:
    :math:`(0, 0, 0, ..., 1, 1, ..., m - 1, m -1)`.
    The second vector is the ``inds_in_sample`` vector.
    This vector is the same length as ``sample_ids``, but its values represent
    the indices of frames within each sample :math:`x_t`.
    For a data set with :math:`T` total frames across all samples,
    where :math:`t_i` indicates the number of frames in each :math:`x_i`,
    this vector will look like :math:`(0, 1, ..., t_0, 0, 1, ..., t_1, ... t_m)`.

    Parameters
    ----------
    subset_df : pandas.DataFrame
        A :class:`pandas.DataFrame` representing the training data subsets.
        This DataFrame is created by
        :func:`vak.prep.frame_classification.learncurve.make_subsets_from_dataset_df`,
        and then passed into this function.
        It is created from  a ``pandas.DataFrame``
        returned by :func:`vak.prep.frame_classification.get_or_make_source_files`
        with a ``'split'`` column added.
    dataset_path : pathlib.Path
        Path to directory that represents dataset.
    input_type : str
        The type of input to the neural network model.
        One of {'audio', 'spect'}.

    Returns
    -------
    None
    """
    subsets = [
        subset for subset in sorted(subsets_df.subset.dropna().unique())
    ]
    for subset in subsets:
        logger.info(f"Making indexing vectors for subset: {subset}")
        subset_df = subsets_df[subsets_df.subset == subset].copy()
        frames_paths = subset_df[
            datapipes.frame_classification.constants.FRAMES_PATH_COL_NAME
        ].values

        def _return_index_arrays(
            source_id_path_tup,
        ):
            """Function we use with dask to parallelize.
            Defined in-line so variables are in scope.
            """
            source_id, frames_path = source_id_path_tup

            frames_path = dataset_path / pathlib.Path(frames_path)

            frames = datapipes.frame_classification.helper.load_frames(
                frames_path, input_type
            )

            n_frames = frames.shape[-1]
            sample_id_vec = np.ones((n_frames,)).astype(np.int32) * source_id
            inds_in_sample_vec = np.arange(n_frames)

            return Sample(
                source_id,
                sample_id_vec,
                inds_in_sample_vec,
            )

        # ---- make npy files for this split, parallelized with dask
        # using nested function just defined
        source_id_frames_path_tups = [
            (source_id, frames_path)
            for source_id, frames_path in enumerate(frames_paths)
        ]

        source_id_frames_path_bag = db.from_sequence(
            source_id_frames_path_tups
        )
        with ProgressBar():
            samples = list(source_id_frames_path_bag.map(_return_index_arrays))
        samples = sorted(samples, key=lambda sample: sample.source_id)

        # ---- save indexing vectors in train directory
        sample_id_vec = np.concatenate(
            list(sample.sample_id_vec for sample in samples)
        )
        np.save(
            dataset_path
            / "train"
            / datapipes.frame_classification.helper.sample_ids_array_filename_for_subset(
                subset
            ),
            sample_id_vec,
        )
        inds_in_sample_vec = np.concatenate(
            list(sample.inds_in_sample_vec for sample in samples)
        )
        np.save(
            dataset_path
            / "train"
            / datapipes.frame_classification.helper.inds_in_sample_array_filename_for_subset(
                subset
            ),
            inds_in_sample_vec,
        )


def make_subsets_from_dataset_df(
    dataset_df: pd.DataFrame,
    input_type: str,
    train_set_durs: Sequence[float],
    num_replicates: int,
    dataset_path: pathlib.Path,
    labelmap: dict,
    background_label: str = common.constants.DEFAULT_BACKGROUND_LABEL,
) -> pd.DataFrame:
    """Make subsets of the training data split for a learning curve.

     Makes subsets given a dataframe representing the entire dataset,
     with one subset for each combination of (training set duration,
     replicate number). Each subset is randomly drawn
     from the total training split.

     Uses :func:`vak.prep.split.frame_classification_dataframe` to make
     subsets of the training data from ``dataset_df``.

     A new column will be added to the dataframe, `'subset'`,
     and additional rows for each subset.
     The dataframe is returned with these subsets added.
     (The `'split'` for these rows will still be `'train'`.)
     Additionally, a separate set of indexing vectors
     will be made for each subset, using
     :func:`vak.prep.frame_classification.learncurve.make_index_vectors_for_each_subset`.

    .. code-block:: console

       032312-vak-frame-classification-dataset-generated-231005_121809
       ├── 032312_prep_231005_121809.csv
       ├── labelmap.json
       ├── metadata.json
       ├── prep_231005_121809.log
       ├── TweetyNet_learncurve_audio_cbin_annot_notmat.toml
       ├── train
           ├── gy6or6_baseline_230312_0808.138.cbin.spect.frame_labels.npy
           ├── gy6or6_baseline_230312_0808.138.cbin.spect.frames.npy
           ├── gy6or6_baseline_230312_0809.141.cbin.spect.frame_labels.npy
           ├── gy6or6_baseline_230312_0809.141.cbin.spect.frames.npy
           ├── gy6or6_baseline_230312_0813.163.cbin.spect.frame_labels.npy
           ├── gy6or6_baseline_230312_0813.163.cbin.spect.frames.npy
           ├── gy6or6_baseline_230312_0816.179.cbin.spect.frame_labels.npy
           ├── gy6or6_baseline_230312_0816.179.cbin.spect.frames.npy
           ├── gy6or6_baseline_230312_0820.196.cbin.spect.frame_labels.npy
           ├── gy6or6_baseline_230312_0820.196.cbin.spect.frames.npy
           ├── inds_in_sample.npy
           ├── inds_in_sample-train-dur-4.0-replicate-1.npy
           ├── inds_in_sample-train-dur-4.0-replicate-2.npy
           ├── inds_in_sample-train-dur-6.0-replicate-1.npy
           ├── inds_in_sample-train-dur-6.0-replicate-2.npy
           ├── sample_ids.npy
           ├── sample_ids-train-dur-4.0-replicate-1.npy
           ├── sample_ids-train-dur-4.0-replicate-2.npy
           ├── sample_ids-train-dur-6.0-replicate-1.npy
           └── sample_ids-train-dur-6.0-replicate-2.npy
       ...

     Parameters
     ----------
     dataset_df : pandas.DataFrame
         Dataframe representing a dataset for frame classification models.
         It is returned by
         :func:`vak.prep.frame_classification.get_or_make_source_files`,
         and has a ``'split'`` column added.
     train_set_durs : list
         Durations in seconds of subsets taken from training data
         to create a learning curve, e.g., `[5., 10., 15., 20.]`.
     num_replicates : int
         number of times to replicate training for each training set duration
         to better estimate metrics for a training set of that size.
         Each replicate uses a different randomly drawn subset of the training
         data (but of the same duration).
     dataset_path : str, pathlib.Path
         Directory where splits will be saved.
     input_type : str
         The type of input to the neural network model.
         One of {'audio', 'spect'}.
    background_label: str, optional
        The string label applied to segments belonging to the
        background class.
        Default is
        :const:`vak.common.constants.DEFAULT_BACKGROUND_LABEL`.

     Returns
     -------
     dataset_df_out : pandas.DataFrame
         A pandas.DataFrame that has the original splits
         from ``dataset_df``, as well as the additional subsets
         of the training data added, along with additional
         columns, ``'subset', 'train_dur', 'replicate_num'``,
         that are used by :mod:`vak`.
         Other functions like :func:`vak.learncurve.learncurve`
         specify a specific subset of the training data
         by getting the subset name with the function
         :func:`vak.common.learncurve.get_train_dur_replicate_split_name`,
         and then filtering ``dataset_df_out`` with that name
         using the 'subset' column.
    """
    dataset_path = pathlib.Path(dataset_path)

    # get just train split, to pass to split.dataframe
    # so we don't end up with other splits in the training set
    train_split_df = dataset_df[dataset_df["split"] == "train"].copy()
    labelset = set([k for k in labelmap.keys() if k != background_label])

    # will concat after loop, then use ``csv_path`` to replace
    # original dataset df with this one
    subsets_df = []
    for train_dur in train_set_durs:
        logger.info(
            f"Subsetting training set for training set of duration: {train_dur}",
        )
        for replicate_num in range(1, num_replicates + 1):
            train_dur_replicate_subset_name = (
                common.learncurve.get_train_dur_replicate_subset_name(
                    train_dur, replicate_num
                )
            )

            train_dur_replicate_df = split.frame_classification_dataframe(
                # copy to avoid mutating original train_split_df
                train_split_df.copy(),
                dataset_path,
                train_dur=train_dur,
                labelset=labelset,
            )
            # remove rows where split set to 'None'
            train_dur_replicate_df = train_dur_replicate_df[
                train_dur_replicate_df.split == "train"
            ]
            # next line, make split name in csv match the split name used for directory in dataset dir
            train_dur_replicate_df["subset"] = train_dur_replicate_subset_name
            train_dur_replicate_df["train_dur"] = train_dur
            train_dur_replicate_df["replicate_num"] = replicate_num
            subsets_df.append(train_dur_replicate_df)

    subsets_df = pd.concat(subsets_df)

    make_index_vectors_for_each_subset(
        subsets_df,
        dataset_path,
        input_type,
    )

    # keep the same validation, test, and total train sets by concatenating them with the train subsets
    dataset_df["subset"] = None  # add column but have it be empty
    dataset_df = pd.concat((subsets_df, dataset_df))
    # We reset the entire index across all splits, instead of repeating indices,
    # and we set drop=False because we don't want to add a new column 'index' or 'level_0'.
    # Need to do this again after calling `make_npy_files_for_each_split` since we just
    # did `pd.concat` with the original dataframe
    dataset_df = dataset_df.reset_index(drop=True)
    return dataset_df



"""tests for vak.config.predict module"""
import pytest

import vak.config.predict


class TestPredictConfig:

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'frames_standardizer_path': '/home/user/results_181014_194418/spect_scaler',
                    'checkpoint_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/bl26lb16/results_200620_164245/TweetyNet/checkpoints/max-val-acc-checkpoint.pt',
                    'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/bl26lb16/results_200620_164245/labelmap.json',
                    'batch_size': 11,
                    'num_workers': 16,
                    'trainer': {'accelerator': 'gpu', 'devices': [0]},
                    'output_dir': './tests/data_for_tests/generated/results/predict/audio_cbin_annot_notmat/TweetyNet',
                    'annot_csv_filename': 'bl26lb16.041912.annot.csv',
                    'model': {
                        'TweetyNet': {
                            'network': {
                                'conv1_filters': 8,
                                'conv1_kernel_size': [3, 3],
                                'conv2_filters': 16,
                                'conv2_kernel_size': [5, 5],
                                'pool1_size': [4, 1],
                                'pool1_stride': [4, 1],
                                'pool2_size': [4, 1],
                                'pool2_stride': [4, 1],
                                'hidden_size': 32
                                },
                                'optimizer': {'lr': 0.001}
                            }
                        },
                        'dataset': {
                            'path': '~/some/path/I/made/up/for/now'
                        },
                    }
            ]
    )
    def test_init(self, config_dict):
        config_dict['model'] = vak.config.ModelConfig.from_config_dict(config_dict['model'])
        config_dict['dataset'] = vak.config.DatasetConfig.from_config_dict(config_dict['dataset'])
        config_dict['trainer'] = vak.config.TrainerConfig(**config_dict['trainer'])

        predict_config = vak.config.PredictConfig(**config_dict)

        assert isinstance(predict_config, vak.config.PredictConfig)

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'frames_standardizer_path': '/home/user/results_181014_194418/spect_scaler',
                    'checkpoint_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/bl26lb16/results_200620_164245/TweetyNet/checkpoints/max-val-acc-checkpoint.pt',
                    'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/bl26lb16/results_200620_164245/labelmap.json',
                    'batch_size': 11,
                    'num_workers': 16,
                    'trainer': {'accelerator': 'gpu', 'devices': [0]},
                    'output_dir': './tests/data_for_tests/generated/results/predict/audio_cbin_annot_notmat/TweetyNet',
                    'annot_csv_filename': 'bl26lb16.041912.annot.csv',
                    'model': {
                        'TweetyNet': {
                            'network': {
                                'conv1_filters': 8,
                                'conv1_kernel_size': [3, 3],
                                'conv2_filters': 16,
                                'conv2_kernel_size': [5, 5],
                                'pool1_size': [4, 1],
                                'pool1_stride': [4, 1],
                                'pool2_size': [4, 1],
                                'pool2_stride': [4, 1],
                                'hidden_size': 32
                                },
                                'optimizer': {'lr': 0.001}
                            }
                        },
                        'dataset': {
                            'path': '~/some/path/I/made/up/for/now'
                        },
                    }
            ]
    )
    def test_from_config_dict(self, config_dict):
        predict_config = vak.config.PredictConfig.from_config_dict(config_dict)

        assert isinstance(predict_config, vak.config.PredictConfig)

    def test_from_config_dict_with_real_config(self, a_generated_predict_config_dict):
        predict_table = a_generated_predict_config_dict["predict"]

        predict_config = vak.config.predict.PredictConfig.from_config_dict(predict_table)

        assert isinstance(predict_config, vak.config.predict.PredictConfig)

    @pytest.mark.parametrize(
            'config_dict, expected_exception',
            [
                # missing 'checkpoint_path', should raise KeyError
                (
                    {
                        'frames_standardizer_path': '/home/user/results_181014_194418/spect_scaler',
                        'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/bl26lb16/results_200620_164245/labelmap.json',
                        'batch_size': 11,
                        'num_workers': 16,
                        'trainer': {'accelerator': 'gpu', 'devices': [0]},
                        'output_dir': './tests/data_for_tests/generated/results/predict/audio_cbin_annot_notmat/TweetyNet',
                        'annot_csv_filename': 'bl26lb16.041912.annot.csv',
                        'model': {
                            'TweetyNet': {
                                'network': {
                                    'conv1_filters': 8,
                                    'conv1_kernel_size': [3, 3],
                                    'conv2_filters': 16,
                                    'conv2_kernel_size': [5, 5],
                                    'pool1_size': [4, 1],
                                    'pool1_stride': [4, 1],
                                    'pool2_size': [4, 1],
                                    'pool2_stride': [4, 1],
                                    'hidden_size': 32
                                    },
                                    'optimizer': {'lr': 0.001}
                                }
                            },
                        'dataset': {
                            'path': '~/some/path/I/made/up/for/now'
                        },
                    },
                    KeyError
                ),
                # missing 'dataset', should raise KeyError
                (
                    {
                        'frames_standardizer_path': '/home/user/results_181014_194418/spect_scaler',
                        'checkpoint_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/bl26lb16/results_200620_164245/TweetyNet/checkpoints/max-val-acc-checkpoint.pt',
                        'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/bl26lb16/results_200620_164245/labelmap.json',
                        'batch_size': 11,
                        'num_workers': 16,
                        'trainer': {'accelerator': 'gpu', 'devices': [0]},
                        'output_dir': './tests/data_for_tests/generated/results/predict/audio_cbin_annot_notmat/TweetyNet',
                        'annot_csv_filename': 'bl26lb16.041912.annot.csv',
                        'model': {
                            'TweetyNet': {
                                'network': {
                                    'conv1_filters': 8,
                                    'conv1_kernel_size': [3, 3],
                                    'conv2_filters': 16,
                                    'conv2_kernel_size': [5, 5],
                                    'pool1_size': [4, 1],
                                    'pool1_stride': [4, 1],
                                    'pool2_size': [4, 1],
                                    'pool2_stride': [4, 1],
                                    'hidden_size': 32
                                    },
                                    'optimizer': {'lr': 0.001}
                                }
                            },
                    },
                    KeyError
                ),
                # missing 'model', should raise KeyError
                (
                    {
                        'frames_standardizer_path': '/home/user/results_181014_194418/spect_scaler',
                        'checkpoint_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/bl26lb16/results_200620_164245/TweetyNet/checkpoints/max-val-acc-checkpoint.pt',
                        'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/bl26lb16/results_200620_164245/labelmap.json',
                        'batch_size': 11,
                        'num_workers': 16,
                        'trainer': {'accelerator': 'gpu', 'devices': [0]},
                        'output_dir': './tests/data_for_tests/generated/results/predict/audio_cbin_annot_notmat/TweetyNet',
                        'annot_csv_filename': 'bl26lb16.041912.annot.csv',
                        'dataset': {
                            'path': '~/some/path/I/made/up/for/now'
                        },
                    },
                    KeyError
                ),
            ]
    )
    def test_from_config_dict_raises(self, config_dict, expected_exception):
        with pytest.raises(expected_exception):
            vak.config.PredictConfig.from_config_dict(config_dict)



"""Helper functions for datasets annotated as sequences."""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import pandas as pd

from ..common import annotation


def where_unlabeled_segments(dataset_df: pd.DataFrame) -> npt.NDArray:
    """Returns a Boolean array that is True where
    a sequence has unlabeled segments
    in a dataset of annotated sequences.

    Parameters
    ----------
    dataset_df : pandas.DataFrame
        A dataframe representing a source dataset of audio signals
        or spectrograms, as returned by
        :func:`vak.prep.audio_dataset.prep_audio_dataset` or
        :func:`vak.prep.spectrogram_dataset.prep_spectrogram_dataset`.

    Returns
    -------
    where_unlabeled : numpy.ndarray
        Vector with Boolean dtype, where a True
        element indicates that the
        annotated sequence indexed by this has
        segments that are unlabeled.
    """
    annots = annotation.from_df(dataset_df)
    durations = dataset_df.duration.values

    has_unlabeled_list = []

    for annot, duration in zip(annots, durations):
        has_unlabeled_list.append(annotation.has_unlabeled(annot, duration))

    return np.array(has_unlabeled_list).astype(bool)


def has_unlabeled_segments(dataset_df: pd.DataFrame) -> bool:
    r"""Returns True if a dataset annotated as sequences
    has segments that are unlabeled in any of the sequences.

    Used to decide whether an additional class needs to be added
    to the set of labels :math:`Y = {y_1, y_2, \dots, y_n}`,
    where the added class :math:`y_{n+1}`
    will represent the unlabeled "background" periods.

    Parameters
    ----------
    dataset_df : pandas.DataFrame
        A dataframe representing a source dataset of audio signals
        or spectrograms, as returned by
        :func:`vak.prep.audio_dataset.prep_audio_dataset` or
        :func:`vak.prep.spectrogram_dataset.prep_spectrogram_dataset`.

    Returns
    -------
    has_unlabeled : bool
        If True, there are annotations in the dataset
        that have unlabeled periods.
    """
    # NOTE we cast to `bool` because
    # np.any returns an instance of <class 'numpy.bool_'>
    # and `<class 'numpy.bool_'> is True == False`.
    # Not sure if this is numpy version specific
    return bool(np.any(where_unlabeled_segments(dataset_df)))



"""Parametric UMAP loss function."""

from __future__ import annotations

import warnings

import torch

# isort: off
# Ignore warnings from Numba deprecation:
# https://numba.readthedocs.io/en/stable/reference/deprecation.html#deprecation-of-object-mode-fall-back-behaviour-when-using-jit
# Numba is required by UMAP.
from numba.core.errors import NumbaDeprecationWarning
from torch.nn.functional import mse_loss

warnings.simplefilter("ignore", category=NumbaDeprecationWarning)
from umap.umap_ import find_ab_params  # noqa : E402

# isort: on


def convert_distance_to_probability(distances, a=1.0, b=1.0):
    """Convert distances to probability.

    Computes equation (2.6) of Sainburg McInnes Gentner 2021,
    :math:`q_{ij} = (1 + a \abs{z_i - z_j}^{2b} )^{-1}`.

    The function uses torch.log1p to avoid floating point error:
    ``-torch.log1p(a * distances ** (2 * b))``.
    See https://en.wikipedia.org/wiki/Natural_logarithm#lnp1
    """
    # next line, equivalent to 1.0 / (1.0 + a * distances ** (2 * b))
    # but avoids floating point error
    return -torch.log1p(a * distances ** (2 * b))


def compute_cross_entropy(
    probabilities_graph,
    probabilities_distance,
    EPS=1e-4,
    repulsion_strength=1.0,
):
    """Computes cross entropy as used for UMAP cost function"""
    # cross entropy
    attraction_term = -probabilities_graph * torch.nn.functional.logsigmoid(
        probabilities_distance
    )
    repulsion_term = (
        -(1.0 - probabilities_graph)
        * (
            torch.nn.functional.logsigmoid(probabilities_distance)
            - probabilities_distance
        )
        * repulsion_strength
    )

    # balance the expected losses between attraction and repulsion
    CE = attraction_term + repulsion_term
    return attraction_term, repulsion_term, CE


def umap_loss(
    embedding_to: torch.Tensor,
    embedding_from: torch.Tensor,
    a,
    b,
    negative_sample_rate: int = 5,
):
    """UMAP loss function

    Converts distances to probabilities,
    and then computes cross entropy.
    """
    # get negative samples by randomly shuffling the batch
    embedding_neg_to = embedding_to.repeat(negative_sample_rate, 1)
    repeat_neg = embedding_from.repeat(negative_sample_rate, 1)
    embedding_neg_from = repeat_neg[torch.randperm(repeat_neg.shape[0])]
    distance_embedding = torch.cat(
        (
            (embedding_to - embedding_from).norm(dim=1),
            (embedding_neg_to - embedding_neg_from).norm(dim=1),
            # ``to`` method in next line to avoid error `Expected all tensors to be on the same device`
        ),
        dim=0,
    ).to(embedding_to.device)

    # convert probabilities to distances
    probabilities_distance = convert_distance_to_probability(
        distance_embedding, a, b
    )
    # set true probabilities based on negative sampling
    batch_size = embedding_to.shape[0]
    probabilities_graph = torch.cat(
        (
            torch.ones(batch_size),
            torch.zeros(batch_size * negative_sample_rate),
        ),
        dim=0,
        # ``to`` method in next line to avoid error `Expected all tensors to be on the same device`
    ).to(embedding_to.device)

    # compute cross entropy
    (attraction_loss, repellant_loss, ce_loss) = compute_cross_entropy(
        probabilities_graph,
        probabilities_distance,
    )
    loss = torch.mean(ce_loss)
    return loss


class UmapLoss(torch.nn.Module):
    """"""

    def __init__(
        self,
        spread: float = 1.0,
        min_dist: float = 0.1,
        negative_sample_rate: int = 5,
        beta: float = 1.0,
    ):
        super().__init__()
        self.min_dist = min_dist
        self.a, self.b = find_ab_params(spread, min_dist)
        self.negative_sample_rate = negative_sample_rate
        self.beta = beta

    def forward(
        self,
        embedding_to: torch.Tensor,
        embedding_from: torch.Tensor,
        reconstruction: torch.Tensor | None = None,
        before_encoding: torch.Tensor | None = None,
    ):
        loss_umap = umap_loss(
            embedding_to,
            embedding_from,
            self.a,
            self.b,
            self.negative_sample_rate,
        )
        if reconstruction is not None:
            loss_reconstruction = mse_loss(reconstruction, before_encoding)
            loss = loss_umap + self.beta * loss_reconstruction
        else:
            loss_reconstruction = None
            loss = loss_umap
        return loss_umap, loss_reconstruction, loss



"""Fixtures used just by test_datasets"""
import json

import numpy as np
import pandas as pd
import pytest


SPLITS_JSON = {
    "splits_csv_path": "splits/inputs-targets-paths-csvs/Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.csv",
    "sample_id_vec_path": {
        "test": "splits/sample-id-vectors/Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.test.sample_ids.npy",
        "train": "splits/sample-id-vectors/Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.train.sample_ids.npy",
        "val": "splits/sample-id-vectors/Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.val.sample_ids.npy"
    },
    "inds_in_sample_vec_path": {
        "test": "splits/inds-in-sample-vectors/Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.test.inds_in_sample.npy",
        "train": "splits/inds-in-sample-vectors/Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.train.inds_in_sample.npy",
        "val": "splits/inds-in-sample-vectors/Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.val.inds_in_sample.npy"
    }
}

INPUTS_TARGETS_CSV_RECORDS = [
    {
        'frames_path': '0.wav.spect.npz',
        'multi_frame_labels_path': '0.wav.multi-frame-labels.npy',
        'binary_frame_labels_path': '0.wav.binary-frame-labels.npy',
        'boundary_frame_labels_path': '0.wav.boundary-frame-labels.npy',
        'split': 'train'
    },
    {
        'frames_path': '1.wav.spect.npz',
        'multi_frame_labels_path': '1.wav.multi-frame-labels.npy',
        'binary_frame_labels_path': '1.wav.binary-frame-labels.npy',
        'boundary_frame_labels_path': '1.wav.boundary-frame-labels.npy',
        'split': 'val'
    },
    {
        'frames_path': '2.wav.spect.npz',
        'multi_frame_labels_path': '2.wav.multi-frame-labels.npy',
        'binary_frame_labels_path': '2.wav.binary-frame-labels.npy',
        'boundary_frame_labels_path': '2.wav.boundary-frame-labels.npy',
        'split': 'test'
    },
    {
        'frames_path': '3.wav.spect.npz',
        'multi_frame_labels_path': '3.wav.multi-frame-labels.npy',
        'binary_frame_labels_path': '3.wav.binary-frame-labels.npy',
        'boundary_frame_labels_path': '3.wav.boundary-frame-labels.npy',
        'split': 'predict'
    },
]

LABELMAPS_JSON = {
    "Bengalese-Finch-Song": {
        "syllable": {
            "bl26lb16": {
                "background": 0,
                "a": 1,
                "b": 2,
                "c": 3,
                "d": 4,
                "e": 5,
                "f": 6,
                "i": 7
            },
            "gr41rd51": {
                "background": 0,
                "a": 1,
                "b": 2,
                "c": 3,
                "d": 4,
                "e": 5,
                "f": 6,
                "g": 7,
                "i": 8,
                "j": 9,
                "k": 10,
                "m": 11
            },
            "gy6or6": {
                "background": 0,
                "a": 1,
                "b": 2,
                "c": 3,
                "d": 4,
                "e": 5,
                "f": 6,
                "g": 7,
                "h": 8,
                "i": 9,
                "j": 10,
                "k": 11
            },
            "or60yw70": {
                "background": 0,
                "a": 1,
                "b": 2,
                "c": 3,
                "d": 4,
                "e": 5,
                "f": 6,
                "g": 7,
                "i": 8
            },
            "Bird0": {
                "background": 0,
                "0": 1,
                "1": 2,
                "2": 3,
                "3": 4,
                "4": 5,
                "5": 6,
                "6": 7,
                "7": 8,
                "8": 9,
                "9": 10
            },
            "Bird4": {
                "background": 0,
                "0": 1,
                "1": 2,
                "2": 3,
                "3": 4,
                "4": 5,
                "5": 6,
                "6": 7,
                "7": 8
            },
            "Bird7": {
                "background": 0,
                "0": 1,
                "1": 2,
                "2": 3,
                "3": 4,
                "4": 5,
                "5": 6,
                "6": 7
            },
            "Bird9": {
                "background": 0,
                "0": 1,
                "1": 2,
                "2": 3,
                "3": 4,
                "4": 5,
                "5": 6
            }
        }
    },
    "Canary-Song": {
        "syllable": {
            "llb3": {
                "background": 0,
                "1": 1,
                "10": 2,
                "11": 3,
                "12": 4,
                "13": 5,
                "14": 6,
                "15": 7,
                "16": 8,
                "17": 9,
                "18": 10,
                "19": 11,
                "2": 12,
                "20": 13,
                "3": 14,
                "4": 15,
                "5": 16,
                "6": 17,
                "7": 18,
                "8": 19,
                "9": 20
            },
            "llb11": {
                "background": 0,
                "1": 1,
                "10": 2,
                "11": 3,
                "12": 4,
                "13": 5,
                "14": 6,
                "15": 7,
                "16": 8,
                "17": 9,
                "18": 10,
                "19": 11,
                "2": 12,
                "20": 13,
                "21": 14,
                "22": 15,
                "23": 16,
                "24": 17,
                "25": 18,
                "26": 19,
                "27": 20,
                "3": 21,
                "4": 22,
                "5": 23,
                "6": 24,
                "7": 25,
                "8": 26,
                "9": 27
            },
            "llb16": {
                "background": 0,
                "1": 1,
                "10": 2,
                "11": 3,
                "12": 4,
                "13": 5,
                "14": 6,
                "15": 7,
                "16": 8,
                "17": 9,
                "18": 10,
                "19": 11,
                "2": 12,
                "20": 13,
                "21": 14,
                "22": 15,
                "23": 16,
                "24": 17,
                "25": 18,
                "26": 19,
                "27": 20,
                "28": 21,
                "29": 22,
                "3": 23,
                "30": 24,
                "4": 25,
                "5": 26,
                "6": 27,
                "7": 28,
                "8": 29,
                "9": 30
            }
        }
    },
    "Human-Speech": {
        "phoneme": {
            "all": {
                "background": 0,
                "aa": 1,
                "ae": 2,
                "ah": 3,
                "ao": 4,
                "aw": 5,
                "ax": 6,
                "ax-h": 7,
                "axr": 8,
                "ay": 9,
                "b": 10,
                "bcl": 11,
                "ch": 12,
                "d": 13,
                "dcl": 14,
                "dh": 15,
                "dx": 16,
                "eh": 17,
                "el": 18,
                "em": 19,
                "en": 20,
                "eng": 21,
                "epi": 22,
                "er": 23,
                "ey": 24,
                "f": 25,
                "g": 26,
                "gcl": 27,
                "h#": 28,
                "hh": 29,
                "hv": 30,
                "ih": 31,
                "ix": 32,
                "iy": 33,
                "jh": 34,
                "k": 35,
                "kcl": 36,
                "l": 37,
                "m": 38,
                "n": 39,
                "ng": 40,
                "nx": 41,
                "ow": 42,
                "oy": 43,
                "p": 44,
                "pau": 45,
                "pcl": 46,
                "q": 47,
                "r": 48,
                "s": 49,
                "sh": 50,
                "t": 51,
                "tcl": 52,
                "th": 53,
                "uh": 54,
                "uw": 55,
                "ux": 56,
                "v": 57,
                "w": 58,
                "y": 59,
                "z": 60,
                "zh": 61
            }
        }
    },
    "Mouse-Pup-Call": {
        "call": {
            "all": {
                "background": 0,
                "BK": 1,
                "BW": 2,
                "GO": 3,
                "LL": 4,
                "LO": 5,
                "MU": 6,
                "MZ": 7,
                "NB": 8,
                "PO": 9,
                "SW": 10
            }
        }
    },
    "Zebra-Finch-Song": {
        "syllable": {
            "blu285": {
                "background": 0,
                "syll_0": 1,
                "syll_1": 2,
                "syll_2": 3,
                "syll_3": 4,
                "syll_4": 5,
                "syll_5": 6
            }
        }
    }
}


@pytest.fixture
def mock_biosoundsegbench_dataset(tmp_path):
    dataset_path = tmp_path / "BioSoundSegBench"
    dataset_path.mkdir()
    splits_dir = dataset_path / "splits"
    splits_dir.mkdir()

    inputs_targets_csv_dir = splits_dir / "inputs-targets-paths-csvs"
    inputs_targets_csv_dir.mkdir()
    df = pd.DataFrame.from_records(INPUTS_TARGETS_CSV_RECORDS)
    splits_csv = df.to_csv(inputs_targets_csv_dir / "Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.csv")
    df.to_csv(splits_csv)

    sample_id_vecs_dir = splits_dir / "sample-id-vectors"
    sample_id_vecs_dir.mkdir()
    inds_in_sample_vecs_dir = splits_dir / "inds-in-sample-vectors"
    inds_in_sample_vecs_dir.mkdir()

    for split in "train", "val", "test":
        sample_id_vec = np.zeros(10)
        np.save(sample_id_vecs_dir / f"Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.{split}.sample_ids.npy", sample_id_vec)
        inds_in_sample_vec = np.arange(10)
        np.save(inds_in_sample_vecs_dir / f"Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.{split}.inds_in_sample.npy", inds_in_sample_vec)

    splits_path = dataset_path / "Mouse-Pup-Call.call.id-SW.frame-dur-1.5-ms.id-data-only.train-dur-1500.0.replicate-1.splits.json"
    with splits_path.open('w') as fp:
        json.dump(SPLITS_JSON, fp)

    with (dataset_path / 'labelmaps.json').open('w') as fp:
        json.dump(LABELMAPS_JSON, fp)

    return dataset_path, splits_path



"""Function that trains models in the frame classification family."""

from __future__ import annotations

import datetime
import json
import logging
import pathlib
import shutil

import lightning
import joblib
import pandas as pd
import torch.utils.data

from .. import datapipes, datasets, models, transforms
from ..common import validators
from ..datapipes.frame_classification import InferDatapipe, TrainDatapipe

logger = logging.getLogger(__name__)


def get_split_dur(df: pd.DataFrame, split: str) -> float:
    """Get duration of a split in a dataset from a pandas DataFrame representing the dataset."""
    return df[df["split"] == split]["duration"].sum()


def get_train_callbacks(
    ckpt_root: str | pathlib.Path,
    ckpt_step: int,
    patience: int,
    checkpoint_monitor: str = "val_acc",
    early_stopping_monitor: str = "val_acc",
    early_stopping_mode: str = "max",
) -> list[lightning.pytorch.callbacks.Callback]:
    ckpt_callback = lightning.pytorch.callbacks.ModelCheckpoint(
        dirpath=ckpt_root,
        filename="checkpoint",
        every_n_train_steps=ckpt_step,
        save_last=True,
        verbose=True,
    )
    ckpt_callback.CHECKPOINT_NAME_LAST = "checkpoint"
    ckpt_callback.FILE_EXTENSION = ".pt"

    val_ckpt_callback = lightning.pytorch.callbacks.ModelCheckpoint(
        monitor=checkpoint_monitor,
        dirpath=ckpt_root,
        save_top_k=1,
        mode="max",
        filename="max-val-acc-checkpoint",
        auto_insert_metric_name=False,
        verbose=True,
    )
    val_ckpt_callback.FILE_EXTENSION = ".pt"

    early_stopping = lightning.pytorch.callbacks.EarlyStopping(
        mode=early_stopping_mode,
        monitor=early_stopping_monitor,
        patience=patience,
        verbose=True,
    )

    return [ckpt_callback, val_ckpt_callback, early_stopping]


def get_trainer(
    accelerator: str,
    devices: int | list[int],
    max_steps: int,
    log_save_dir: str | pathlib.Path,
    val_step: int,
    callback_kwargs: dict | None = None,
) -> lightning.pytorch.Trainer:
    """Returns an instance of :class:`lightning.pytorch.Trainer`
    with a default set of callbacks.

    Used by :func:`vak.train.frame_classification`.
    The default set of callbacks is provided by
    :func:`get_default_train_callbacks`.

    Parameters
    ----------
    accelerator : str
    devices : int, list of int
    max_steps : int
    log_save_dir : str, pathlib.Path
    val_step : int
    default_callback_kwargs : dict, optional

    Returns
    -------
    trainer : lightning.pytorch.Trainer

    """
    if callback_kwargs:
        callbacks = get_train_callbacks(**callback_kwargs)
    else:
        callbacks = None

    logger = lightning.pytorch.loggers.TensorBoardLogger(save_dir=log_save_dir)

    trainer = lightning.pytorch.Trainer(
        accelerator=accelerator,
        devices=devices,
        callbacks=callbacks,
        val_check_interval=val_step,
        max_steps=max_steps,
        logger=logger,
    )
    return trainer


def train_frame_classification_model(
    model_config: dict,
    dataset_config: dict,
    trainer_config: dict,
    batch_size: int,
    num_epochs: int,
    num_workers: int,
    checkpoint_path: str | pathlib.Path | None = None,
    frames_standardizer_path: str | pathlib.Path | None = None,
    results_path: str | pathlib.Path | None = None,
    standardize_frames: bool = True,
    shuffle: bool = True,
    val_step: int | None = None,
    ckpt_step: int | None = None,
    patience: int | None = None,
    subset: str | None = None,
) -> None:
    """Train a model from the frame classification family
    and save results.

    Saves checkpoint files for model,
    label map, and spectrogram scaler.
    These are saved either in ``results_path``
    if specified, or a new directory
    made inside ``root_results_dir``.

    Parameters
    ----------
    model_config : dict
        Model configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.ModelConfig.asdict`.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer` in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    batch_size : int
        number of samples per batch presented to models during training.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader.
        Optional, default is None.
    checkpoint_path : str, pathlib.Path
        path to a checkpoint file,
        e.g., one generated by a previous run of ``vak.core.train``.
        If specified, this checkpoint will be loaded into model.
        Used when continuing training.
        Default is None, in which case a new model is initialized.
    frames_standardizer_path : str, pathlib.Path
        path to a saved :class:`~vak.transforms.FramesStandardizer`
        used to standardize (normalize) frames, the input to a
        frame classification model.
        e.g., one generated by a previous run of :func:`vak.core.train`.
        Used when continuing training, for example on the same dataset.
        Default is None.
    root_results_dir : str, pathlib.Path
        Root directory in which a new directory will be created
        where results will be saved.
    results_path : str, pathlib.Path
        Directory where results will be saved.
        If specified, this parameter overrides ``root_results_dir``.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    standardize_frames : bool
        if True, use :class:`vak.transforms.FramesStandardizer` to standardize the frames.
        Normalization is done by subtracting off the mean for each row
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    val_step : int
        Step on which to estimate accuracy using validation set.
        If val_step is n, then validation is carried out every time
        the global step / n is a whole number, i.e., when val_step modulo the global step is 0.
        Default is None, in which case no validation is done.
    ckpt_step : int
        Step on which to save to checkpoint file.
        If ckpt_step is n, then a checkpoint is saved every time
        the global step / n is a whole number, i.e., when ckpt_step modulo the global step is 0.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of validation steps to wait without performance on the
        validation set improving before stopping the training.
        Default is None, in which case training only stops after the specified number of epochs.
    subset : str
        Name of a subset from the training split of the dataset
        to use when training model. This parameter is used by
        :func:`vak.learncurve.learncurve` to specify subsets
        when training models for a learning curve.
    """
    for path, path_name in zip(
        (checkpoint_path, frames_standardizer_path),
        ("checkpoint_path", "frames_standardizer_path"),
    ):
        if path is not None:
            if not validators.is_a_file(path):
                raise FileNotFoundError(
                    f"value for ``{path_name}`` not recognized as a file: {path}"
                )

    model_name = model_config["name"]  # we use this var again below
    if "window_size" not in dataset_config["params"]:
        raise KeyError(
            f"The `dataset_config` for frame classification model '{model_name}' must include a 'params' sub-table "
            f"that sets a value for 'window_size', but received a `dataset_config` that did not:\n{dataset_config}"
        )

    dataset_path = pathlib.Path(dataset_config["path"])
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    # we do this first to make sure we can save things in `results_path`: copy of toml config file, labelset.json, etc
    results_path = pathlib.Path(results_path).expanduser().resolve()
    if not results_path.is_dir():
        raise NotADirectoryError(
            f"`results_path` not recognized as a directory: {results_path}"
        )
    logger.info(
        f"Will save results in `results_path`: {results_path}",
    )

    logger.info(
        f"Loading dataset from `dataset_path`: {dataset_path}\nUsing dataset config: {dataset_config}"
    )
    # ---------------- load training data  -----------------------------------------------------------------------------
    # ---- *not* using a built-in dataset ------------------------------------------------------------------------------
    if dataset_config["name"] is None:
        metadata = datapipes.frame_classification.Metadata.from_dataset_path(
            dataset_path
        )
        dataset_csv_path = dataset_path / metadata.dataset_csv_filename
        dataset_df = pd.read_csv(dataset_csv_path)
        # we have to check this pre-condition here since we need `dataset_df` to check
        if val_step and not dataset_df["split"].str.contains("val").any():
            raise ValueError(
                f"val_step set to {val_step} but dataset does not contain a validation set; "
                f"please run `vak prep` with a config.toml file that specifies a duration for the validation set."
            )

        frame_dur = metadata.frame_dur
        logger.info(
            f"Duration of a frame in dataset, in seconds: {frame_dur}",
        )

        logger.info(f"Using training split from dataset: {dataset_path}")
        train_dur = get_split_dur(dataset_df, "train")
        logger.info(
            f"Total duration of training split from dataset (in s): {train_dur}",
        )

        labelmap_path = dataset_path / "labelmap.json"
        logger.info(f"loading labelmap from path: {labelmap_path}")
        with labelmap_path.open("r") as f:
            labelmap = json.load(f)
        # copy to new results_path
        with open(results_path.joinpath("labelmap.json"), "w") as f:
            json.dump(labelmap, f)

        if frames_standardizer_path is not None and standardize_frames:
            logger.info(
                f"Loading frames standardizer from path: {frames_standardizer_path}"
            )
            frames_standardizer = joblib.load(frames_standardizer_path)
            shutil.copy(frames_standardizer_path, results_path)
        # get transforms just before creating datasets with them
        elif standardize_frames and frames_standardizer_path is None:
            logger.info(
                "No `frames_standardizer_path` provided, not loading",
            )
            logger.info("Will standardize (normalize) frames")
            frames_standardizer = (
                transforms.FramesStandardizer.fit_dataset_path(
                    dataset_path,
                    split="train",
                    subset=subset,
                )
            )
            joblib.dump(
                frames_standardizer,
                results_path.joinpath("FramesStandardizer"),
            )
        elif frames_standardizer_path is not None and not standardize_frames:
            raise ValueError(
                "`frames_standardizer_path` provided but `standardize_frames` was False, these options conflict"
            )
        # ---- *yes* using a built-in dataset --------------------------------------------------------------------------
        else:
            # not standardize_frames and frames_standardizer_path is None:
            logger.info(
                "`standardize_frames` is False and no `frames_standardizer_path` was provided, "
                "will not standardize spectrograms",
            )
            frames_standardizer = None

        train_dataset = TrainDatapipe.from_dataset_path(
            dataset_path=dataset_path,
            split="train",
            subset=subset,
            window_size=dataset_config["params"]["window_size"],
            frames_standardizer=frames_standardizer,
        )
    else:
        # ---- we are using a built-in dataset -----------------------------------------
        # TODO: fix this hack
        # (by doing the same thing with the built-in datapipes, making this a Boolean parameter
        # while still accepting a transform but defaulting to None)
        if "standardize_frames" not in dataset_config:
            logger.info(
                f'Adding `standardize_frames` argument to dataset_config["params"]: {standardize_frames}'
            )
            dataset_config["params"]["standardize_frames"] = standardize_frames
        train_dataset = datasets.get(
            dataset_config,
            split="train",
        )
        frame_dur = train_dataset.frame_dur
        logger.info(
            f"Duration of a frame in dataset, in seconds: {frame_dur}",
        )
        # copy labelmap from dataset to new results_path
        labelmap = train_dataset.labelmap
        with open(results_path.joinpath("labelmap.json"), "w") as fp:
            json.dump(labelmap, fp)
        frames_standardizer = getattr(
            train_dataset.item_transform, "frames_standardizer"
        )
        if frames_standardizer is not None:
            logger.info(
                "Saving `frames_standardizer` from item transform on training dataset"
            )
            joblib.dump(
                frames_standardizer,
                results_path.joinpath("FramesStandardizer"),
            )

    logger.info(
        f"Duration of {train_dataset.__class__.__name__} used for training, in seconds: {train_dataset.duration}",
    )
    train_loader = torch.utils.data.DataLoader(
        dataset=train_dataset,
        shuffle=shuffle,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    # ---------------- load validation set (if there is one) -----------------------------------------------------------
    if val_step:
        logger.info(
            f"Will measure error on validation set every {val_step} steps of training",
        )
        if dataset_config["name"] is None:
            logger.info(
                f"Using validation split from dataset:\n{dataset_path}"
            )
            val_dur = get_split_dur(dataset_df, "val")
            logger.info(
                f"Total duration of validation split from dataset (in s): {val_dur}",
            )
            val_dataset = InferDatapipe.from_dataset_path(
                dataset_path=dataset_path,
                split="val",
                **dataset_config["params"],
                frames_standardizer=frames_standardizer,
                return_padding_mask=True,
            )
        else:
            dataset_config["params"]["return_padding_mask"] = True
            val_dataset = datasets.get(
                dataset_config,
                split="val",
                frames_standardizer=frames_standardizer,
            )
        logger.info(
            f"Duration of {val_dataset.__class__.__name__} used for evaluation, in seconds: {val_dataset.duration}",
        )
        val_loader = torch.utils.data.DataLoader(
            dataset=val_dataset,
            shuffle=False,
            # batch size 1 because each spectrogram reshaped into a batch of windows
            batch_size=1,
            num_workers=num_workers,
        )
    else:
        val_loader = None

    model = models.get(
        model_name,
        model_config,
        num_classes=len(labelmap),
        input_shape=train_dataset.shape,
        labelmap=labelmap,
    )

    if checkpoint_path is not None:
        logger.info(
            f"loading checkpoint for {model_name} from path: {checkpoint_path}",
        )
        model.load_state_dict_from_path(checkpoint_path)

    results_model_root = results_path.joinpath(model_name)
    results_model_root.mkdir()
    ckpt_root = results_model_root.joinpath("checkpoints")
    ckpt_root.mkdir()
    logger.info(f"training {model_name}")
    max_steps = num_epochs * len(train_loader)
    if "target_type" in dataset_config["params"]:
        if isinstance(dataset_config["params"]["target_type"], list) and all([isinstance(target_type, str) for target_type in dataset_config["params"]["target_type"]]):
            multiple_targets = True
        elif isinstance(dataset_config["params"]["target_type"], str):
            multiple_targets = False
        else:
            raise ValueError(
                f'Invalid value for dataset_config["params"]["target_type"]: {dataset_config["params"]["target_type"], list}'
            )
    else:
        multiple_targets = False

    callback_kwargs = dict(
        ckpt_root=ckpt_root,
        ckpt_step=ckpt_step,
        patience=patience,
        checkpoint_monitor="val_multi_acc" if multiple_targets else "val_acc",
        early_stopping_monitor="val_multi_acc" if multiple_targets else "val_acc",
        early_stopping_mode="max",
    )
    trainer = get_trainer(
        accelerator=trainer_config["accelerator"],
        devices=trainer_config["devices"],
        max_steps=max_steps,
        log_save_dir=results_model_root,
        val_step=val_step,
        callback_kwargs=callback_kwargs,
    )
    train_time_start = datetime.datetime.now()
    logger.info(f"Training start time: {train_time_start.isoformat()}")
    trainer.fit(
        model=model,
        train_dataloaders=train_loader,
        val_dataloaders=val_loader,
    )
    train_time_stop = datetime.datetime.now()
    logger.info(f"Training stop time: {train_time_stop.isoformat()}")
    elapsed = train_time_stop - train_time_start
    logger.info(f"Elapsed training time: {elapsed}")



"""Function that evaluates trained models in the frame classification family."""

from __future__ import annotations

import json
import logging
import pathlib
from collections import OrderedDict
from datetime import datetime

import joblib
import lightning
import pandas as pd
import torch.utils.data

from .. import datapipes, datasets, models, transforms
from ..common import constants, validators
from ..datapipes.frame_classification import InferDatapipe

logger = logging.getLogger(__name__)


def eval_frame_classification_model(
    model_config: dict,
    dataset_config: dict,
    trainer_config: dict,
    checkpoint_path: str | pathlib.Path,
    labelmap_path: str | pathlib.Path,
    output_dir: str | pathlib.Path,
    num_workers: int,
    frames_standardizer_path: str | pathlib.Path = None,
    post_tfm_kwargs: dict | None = None,
) -> None:
    """Evaluate a trained model.

    Parameters
    ----------
    model_config : dict
        Model configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.ModelConfig.asdict`.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    checkpoint_path : str, pathlib.Path
        Path to directory with checkpoint files saved by Torch, to reload model
    output_dir : str, pathlib.Path
        Path to location where .csv files with evaluation metrics should be saved.
    labelmap_path : str, pathlib.Path
        Path to 'labelmap.json' file.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    frames_standardizer_path : str, pathlib.Path
        Path to a saved :class:`vak.transforms.FramesStandardizer`
        object used to standardize (normalize) frames.
        If frames were standardized during training and this is not provided,
        will give incorrect results. Default is None.
    post_tfm_kwargs : dict
        Keyword arguments to post-processing transform.
        If None, then no additional clean-up is applied
        when transforming labeled timebins to segments,
        the default behavior. The transform used is
        ``vak.transforms.frame_labels.PostProcess`.
        Valid keyword argument names are 'majority_vote'
        and 'min_segment_dur', and should be appropriate
        values for those arguments: Boolean for ``majority_vote``,
        a float value for ``min_segment_dur``.
        See the docstring of the transform for more details on
        these arguments and how they work.

    Notes
    -----
    Note that unlike :func:`~vak.predict.predict`, this function
    can modify ``labelmap`` so that metrics like edit distance
    are correctly computed, by converting any string labels
    in ``labelmap`` with multiple characters
    to (mock) single-character labels,
    with :func:`vak.labels.multi_char_labels_to_single_char`.
    """
    # ---- pre-conditions ----------------------------------------------------------------------------------------------
    for path, path_name in zip(
        (checkpoint_path, labelmap_path, frames_standardizer_path),
        ("checkpoint_path", "labelmap_path", "frames_standardizer_path"),
    ):
        if path is not None:  # because `frames_standardizer_path` is optional
            if not validators.is_a_file(path):
                raise FileNotFoundError(
                    f"value for ``{path_name}`` not recognized as a file: {path}"
                )

    model_name = model_config["name"]  # we use this var again below
    if "window_size" not in dataset_config["params"]:
        raise KeyError(
            f"The `dataset_config` for frame classification model '{model_name}' must include a 'params' sub-table "
            f"that sets a value for 'window_size', but received a `dataset_config` that did not:\n{dataset_config}"
        )

    if not validators.is_a_directory(output_dir):
        raise NotADirectoryError(
            f"value for ``output_dir`` not recognized as a directory: {output_dir}"
        )

    # ---- get time for .csv file --------------------------------------------------------------------------------------
    timenow = datetime.now().strftime("%y%m%d_%H%M%S")

    # ---- load what we need to transform data -------------------------------------------------------------------------
    if frames_standardizer_path:
        logger.info(
            f"loading frames standardizer from path: {frames_standardizer_path}"
        )
        frames_standardizer = joblib.load(frames_standardizer_path)
    else:
        logger.info(
            "No `frames_standardizer_path` provided, not standardizing frames."
        )
        frames_standardizer = None

    logger.info(f"loading labelmap from path: {labelmap_path}")
    with labelmap_path.open("r") as f:
        labelmap = json.load(f)

    # ---------------- load data for evaluation ------------------------------------------------------------------------
    if "split" in dataset_config["params"]:
        split = dataset_config["params"]["split"]
        # we do this convoluted thing to avoid 'TypeError: Dataset got multiple values for split`
        del dataset_config["params"]["split"]
    else:
        split = "test"
    # ---- *not* using a built-in dataset ------------------------------------------------------------------------------
    if dataset_config["name"] is None:
        dataset_path = pathlib.Path(dataset_config["path"])
        if not dataset_path.exists() or not dataset_path.is_dir():
            raise NotADirectoryError(
                f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
            )

        # we unpack `frame_dur` to log it, regardless of whether we use it with post_tfm below
        metadata = datapipes.frame_classification.Metadata.from_dataset_path(
            dataset_path
        )
        frame_dur = metadata.frame_dur
        logger.info(
            f"Duration of a frame in dataset, in seconds: {frame_dur}",
        )
        val_dataset = InferDatapipe.from_dataset_path(
            dataset_path=dataset_path,
            split=split,
            window_size=dataset_config["params"]["window_size"],
            frames_standardizer=frames_standardizer,
            return_padding_mask=True,
        )
    # ---- *yes* using a built-in dataset ------------------------------------------------------------------------------
    else:
        # next line, we don't use dataset path in this code block,
        # but we need it below when we build the DataFrame with eval results.
        # we're unpacking it here just as we do above with a prep'd dataset
        dataset_path = pathlib.Path(dataset_config["path"])
        dataset_config["params"]["return_padding_mask"] = True
        val_dataset = datasets.get(
            dataset_config,
            split=split,
            frames_standardizer=frames_standardizer,
        )
        frame_dur = val_dataset.frame_dur
        logger.info(
            f"Duration of a frame in dataset, in seconds: {frame_dur}",
        )

    val_loader = torch.utils.data.DataLoader(
        dataset=val_dataset,
        shuffle=False,
        # batch size 1 because each spectrogram reshaped into a batch of windows
        batch_size=1,
        num_workers=num_workers,
    )

    # ---------------- do the actual evaluating ------------------------------------------------------------------------
    input_shape = val_dataset.shape
    # if dataset returns spectrogram reshaped into windows,
    # throw out the window dimension; just want to tell network (channels, height, width) shape
    if len(input_shape) == 4:
        input_shape = input_shape[1:]

    if post_tfm_kwargs:
        post_tfm = transforms.frame_labels.PostProcess(
            timebin_dur=frame_dur,
            background_label=labelmap[constants.DEFAULT_BACKGROUND_LABEL],
            **post_tfm_kwargs,
        )
    else:
        post_tfm = None

    model = models.get(
        model_name,
        model_config,
        num_classes=len(labelmap),
        input_shape=input_shape,
        labelmap=labelmap,
        post_tfm=post_tfm,
    )

    logger.info(f"running evaluation for model: {model_name}")

    model.load_state_dict_from_path(checkpoint_path)

    trainer_logger = lightning.pytorch.loggers.TensorBoardLogger(
        save_dir=output_dir
    )
    trainer = lightning.pytorch.Trainer(
        accelerator=trainer_config["accelerator"],
        devices=trainer_config["devices"],
        logger=trainer_logger,
    )
    # TODO: check for hasattr(model, test_step) and if so run test
    # below, [0] because validate returns list of dicts, length of no. of val loaders
    metric_vals = trainer.validate(model, dataloaders=val_loader)[0]
    metric_vals = {f"avg_{k}": v for k, v in metric_vals.items()}
    for metric_name, metric_val in metric_vals.items():
        if metric_name.startswith("avg_"):
            logger.info(f"{metric_name}: {metric_val:0.5f}")

    # create a "DataFrame" with just one row which we will save as a csv;
    # the idea is to be able to concatenate csvs from multiple runs of eval
    row = OrderedDict(
        [
            ("model_name", model_name),
            ("checkpoint_path", checkpoint_path),
            ("labelmap_path", labelmap_path),
            ("frames_standardizer_path", frames_standardizer_path),
            ("dataset_path", dataset_path),
        ]
    )
    # TODO: is this still necessary after switching to Lightning? Stop saying "average"?
    # order metrics by name to be extra sure they will be consistent across runs
    row.update(
        sorted(
            [(k, v) for k, v in metric_vals.items() if k.startswith("avg_")]
        )
    )

    # pass index into dataframe, needed when using all scalar values (a single row)
    # throw away index below when saving to avoid extra column
    eval_df = pd.DataFrame(row, index=[0])
    eval_csv_path = output_dir.joinpath(f"eval_{model_name}_{timenow}.csv")
    logger.info(f"saving csv with evaluation metrics at: {eval_csv_path}")
    eval_df.to_csv(
        eval_csv_path, index=False
    )  # index is False to avoid having "Unnamed: 0" column when loading



"""Function that evaluates trained models in the parametric UMAP family."""

from __future__ import annotations

import logging
import pathlib
from collections import OrderedDict
from datetime import datetime

import lightning
import pandas as pd
import torch.utils.data

from .. import models
from ..common import validators
from ..datapipes.parametric_umap import Datapipe

logger = logging.getLogger(__name__)


def eval_parametric_umap_model(
    model_config: dict,
    dataset_config: dict,
    checkpoint_path: str | pathlib.Path,
    output_dir: str | pathlib.Path,
    batch_size: int,
    num_workers: int,
    trainer_config: dict,
) -> None:
    """Evaluate a trained model.

    Parameters
    ----------
    model_config : dict
        Model configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.ModelConfig.asdict`.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    checkpoint_path : str, pathlib.Path
        Path to directory with checkpoint files saved by Torch, to reload model
    output_dir : str, pathlib.Path
        Path to location where .csv files with evaluation metrics should be saved.
    batch_size : int
        Number of samples per batch fed into model.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    split : str
        Split of dataset on which model should be evaluated.
        One of {'train', 'val', 'test'}. Default is 'test'.
    """
    # ---- pre-conditions ----------------------------------------------------------------------------------------------
    for path, path_name in zip(
        (checkpoint_path,),
        ("checkpoint_path",),
    ):
        if path is not None:  # because `frames_standardizer_path` is optional
            if not validators.is_a_file(path):
                raise FileNotFoundError(
                    f"value for ``{path_name}`` not recognized as a file: {path}"
                )

    dataset_path = pathlib.Path(dataset_config["path"])
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )
    logger.info(
        f"Loading metadata from dataset path: {dataset_path}",
    )

    if not validators.is_a_directory(output_dir):
        raise NotADirectoryError(
            f"value for ``output_dir`` not recognized as a directory: {output_dir}"
        )

    # ---- get time for .csv file --------------------------------------------------------------------------------------
    timenow = datetime.now().strftime("%y%m%d_%H%M%S")

    # ---------------- load data for evaluation ------------------------------------------------------------------------
    if "split" in dataset_config["params"]:
        split = dataset_config["params"]["split"]
    else:
        split = "test"
    model_name = model_config["name"]
    val_dataset = Datapipe.from_dataset_path(
        dataset_path=dataset_path,
        split=split,
        **dataset_config["params"],
    )
    val_loader = torch.utils.data.DataLoader(
        dataset=val_dataset,
        shuffle=False,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    # ---------------- do the actual evaluating ------------------------------------------------------------------------
    model = models.get(
        model_name,
        model_config,
        input_shape=val_dataset.shape,
    )

    logger.info(f"running evaluation for model: {model_name}")

    model.load_state_dict_from_path(checkpoint_path)

    trainer_logger = lightning.pytorch.loggers.TensorBoardLogger(
        save_dir=output_dir
    )
    trainer = lightning.pytorch.Trainer(
        accelerator=trainer_config["accelerator"],
        devices=trainer_config["devices"],
        logger=trainer_logger,
    )
    # TODO: check for hasattr(model, test_step) and if so run test
    # below, [0] because validate returns list of dicts, length of no. of val loaders
    metric_vals = trainer.validate(model, dataloaders=val_loader)[0]
    for metric_name, metric_val in metric_vals.items():
        logger.info(f"{metric_name}: {metric_val:0.5f}")

    # create a "DataFrame" with just one row which we will save as a csv;
    # the idea is to be able to concatenate csvs from multiple runs of eval
    row = OrderedDict(
        [
            ("model_name", model_name),
            ("checkpoint_path", checkpoint_path),
            ("dataset_path", dataset_path),
        ]
    )
    # order metrics by name to be extra sure they will be consistent across runs
    row.update(sorted([(k, v) for k, v in metric_vals.items()]))

    # pass index into dataframe, needed when using all scalar values (a single row)
    # throw away index below when saving to avoid extra column
    eval_df = pd.DataFrame(row, index=[0])
    eval_csv_path = output_dir.joinpath(f"eval_{model_name}_{timenow}.csv")
    logger.info(f"saving csv with evaluation metrics at: {eval_csv_path}")
    eval_df.to_csv(
        eval_csv_path, index=False
    )  # index is False to avoid having "Unnamed: 0" column when loading



"""Class that represents ``[vak.learncurve]`` table in configuration file."""

from __future__ import annotations

from attrs import converters, define, field, validators

from .dataset import DatasetConfig
from .eval import are_valid_post_tfm_kwargs, convert_post_tfm_kwargs
from .model import ModelConfig
from .train import TrainConfig
from .trainer import TrainerConfig

REQUIRED_KEYS = (
    "dataset",
    "model",
    "root_results_dir",
    "trainer",
)


@define
class LearncurveConfig(TrainConfig):
    """Class that represents ``[vak.learncurve]`` table in configuration file.

    Attributes
    ----------
    model : vak.config.ModelConfig
        The model to use: its name,
        and the parameters to configure it.
        Must be an instance of :class:`vak.config.ModelConfig`
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    batch_size : int
        number of samples per batch presented to models during training.
    root_results_dir : str
        directory in which results will be created.
        The vak.cli.train function will create
        a subdirectory in this directory each time it runs.
    dataset : vak.config.DatasetConfig
        The dataset to use: the path to it,
        and optionally a path to a file representing splits,
        and the name, if it is a built-in dataset.
        Must be an instance of :class:`vak.config.DatasetConfig`.
    trainer : vak.config.TrainerConfig
        Configuration for :class:`lightning.pytorch.Trainer`.
        Must be an instance of :class:`vak.config.TrainerConfig`.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    standardize_frames : bool
        if True, use :class:`vak.transforms.FramesStandardizer` to standardize the frames.
        Normalization is done by subtracting off the mean for each row
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    val_step : int
        Step on which to estimate accuracy using validation set.
        If val_step is n, then validation is carried out every time
        the global step / n is a whole number, i.e., when val_step modulo the global step is 0.
        Default is None, in which case no validation is done.
    ckpt_step : int
        step/epoch at which to save to checkpoint file.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of epochs to wait without the error dropping before stopping the
        training. Default is None, in which case training continues for num_epochs
    post_tfm_kwargs : dict
        Keyword arguments to post-processing transform.
        If None, then no additional clean-up is applied
        when transforming labeled timebins to segments,
        the default behavior.
        The transform used is
        ``vak.transforms.frame_labels.ToSegmentsWithPostProcessing`.
        Valid keyword argument names are 'majority_vote'
        and 'min_segment_dur', and should be appropriate
        values for those arguments: Boolean for ``majority_vote``,
        a float value for ``min_segment_dur``.
        See the docstring of the transform for more details on
        these arguments and how they work.
    """

    post_tfm_kwargs = field(
        validator=validators.optional(are_valid_post_tfm_kwargs),
        converter=converters.optional(convert_post_tfm_kwargs),
        default=None,
    )

    # we over-ride this method from TrainConfig mainly so the docstring is correct.
    # TODO: can we do this by just over-writing `__doc__` for the method on this class?
    @classmethod
    def from_config_dict(cls, config_dict: dict) -> LearncurveConfig:
        """Return :class:`LearncurveConfig` instance from a :class:`dict`.

        The :class:`dict` passed in should be the one found
        by loading a valid configuration toml file with
        :func:`vak.config.parse.from_toml_path`,
        and then using key ``learncurve``,
        i.e., ``LearncurveConfig.from_config_dict(config_dict['learncurve'])``.
        """
        for required_key in REQUIRED_KEYS:
            if required_key not in config_dict:
                raise KeyError(
                    "The `[vak.learncurve]` table in a configuration file requires "
                    f"the option '{required_key}', but it was not found "
                    "when loading the configuration file into a Python dictionary. "
                    "Please check that the configuration file is formatted correctly."
                )
        config_dict["model"] = ModelConfig.from_config_dict(
            config_dict["model"]
        )
        config_dict["dataset"] = DatasetConfig.from_config_dict(
            config_dict["dataset"]
        )
        config_dict["trainer"] = TrainerConfig(**config_dict["trainer"])
        return cls(**config_dict)



"""This is the script run by nox session "test-data-generate".

It makes all the 'generated' test data, i.e. files created by vak,
It's called 'generated' test data to distinguish it from the
'source' test data, i.e., files *not* created by vak:
the input data such as audio and annotation files that are used
when vak *does* create files (csv files, logs,
neural network checkpoints, etc.).
This script generates those files for use by unit and integration tests.
Specifically, it generates:
* temporary config.toml files used when generating results
* `prep`d (prepared) datasets, and results created with those datasets,
  both of which were generated using the temporary config.toml files

All the setup configs send output to one of two places:
- for any prep command, the output goes to some child directory of ./tests/data_for_tests/generated/prep
- for any command run with a `prep`d dataset, the output goes to some child directory
  of ./tests/data_for_tests/generated/results

Examples:
    when we run `vak prep tests/data_for_tests/generated/configs/test_train_audio_wav_annot_birdsongrec.toml`
    the `prep`d dataset will be in a new directory created in
    `./tests/data_for_tests/generated/prep/train/audio_wav_annot_birdsongrec`

    when we run `vak train tests/data_for_tests/genereated/configs/test_train_audio_wav_annot_birdsongrec.toml`
    it will use the `prep`d dataset csv that's now in
    `./tests/data_for_tests/generated/prep/train/audio_wav_annot_birdsongrec`, that the temporary config.toml points to,
    and the results will go to a new directory created in
    `./tests/data_for_tests/generated/results/train/audio_wav_annot_birdsongrec`

To set up this directory structure, we iterate through the constants defined below.

The constants are:
- TOP_LEVEL_DIRS
    name of sub-directories in .tests/data_for_tests/generated that correspond to
    either tempory config files, 'prep'd datasets, or `results` generated from those `prep`d datasets
- COMMAND_DIRS
    names of sub-sub-directories in ./tests/data_for_tests/generated/prep &
    ./tests/data_for_tests/generated/results that correspond to cli commands;
    e.g., dataset from running `vak prep $A_TRAIN_CONFIG.TOML` will be in
    ./tests/data_for_tests/generated/prep/train/audio_{}_annot_{}
- DATA_DIRS
    names of "subsub"directories in ./tests/data_for_tests/$COMMAND that correspond to toy test data sets;
    those sub-directories contain test data generated by $COMMAND using the specified toy test data set

In other words, the parent directory for all the generated directories we need to remove
will have a path of the form: `./tests/data_for_tests/$TOP_LEVEL_DIRS/COMMAND_DIRS/DATA_DIRS`.
For example:
The prep`d dataset from running `vak prep $AUDIO_CBIN_ANNOT_NOTMAT_CONFIG.TOML` will be in
`./tests/data_for_tests/prep/train/audio_cbin_annot_notmat`
and corresponding results will be in
`./tests/data_for_tests/results/train/audio_cbin_annot_notmat`
The directories will have names with timestamps like `prep_20201015_1115`.
Those are the generated directories we want to remove.
"""
import logging
import sys
import warnings

# Do this here to suppress warnings before we import vak
from numba.core.errors import NumbaDeprecationWarning
warnings.simplefilter('ignore', category=NumbaDeprecationWarning)
import vak

import vaktestdata


logger = logging.getLogger('vaktestdata')  # 'base' logger
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
stream_handler = logging.StreamHandler(sys.stdout)
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)
logger.setLevel('INFO')


def generate_test_data(
        step: str = 'all',
        commands=vaktestdata.constants.COMMANDS,
        single_train_result:bool = True,
):
    """Main function that generates all the test data.

    Parameters
    ----------
    step : str
        The step of generating test data to carry out.
        One of ('prep', 'results', or 'all').
        This option is provided through the command-line
        because generating all the data sets is time-consuming
        and may not be needed every time test data is re-generated.
        Running the 'prep' step will run just `vak prep`
        for every configuration file.
        Running the 'results' step assumes that the 'prep' step
        has already been run, and it will run the appropriate command
        for each configuration file to generate results with the prep'd
        dataset, e.g. `vak train`.
        Running 'both' will first run 'prep' and then 'results'--this is
        the default.
    commands : list
        List of commands to run for 'results' step.
    """
    # need to run `prep` before we run other commands
    if step in ('prep', 'all'):
        # first we generate outputs of processing steps
        # leading up to a dataset that speed up tests
        vaktestdata.dirs.make_spect_output_dir_in_generated()  # for any prepared spectrograms
        vaktestdata.dirs.make_source_files_csv_dir_in_generated()  # for csvs of source files
        vaktestdata.dirs.make_source_files_with_splits_csv_dir_in_generated()  # same csvs, with splits added

        # -- now actually run prep for all the configs
        config_paths = vaktestdata.configs.copy_config_files()

        vaktestdata.source_files.set_up_source_files_and_csv_files_for_frame_classification_models()

        vaktestdata.dirs.make_subdirs_in_generated(config_paths)
        # run prep for some models
        vaktestdata.prep.run_prep()
        # re-use some of the prepped datasets for other models
        # this makes time to prep all datasets shorter
        vaktestdata.configs.add_dataset_path_from_prepped_configs()

    if step in ('results', 'all'):
        # Note we need to run `train` first to get results needed for `eval', 'predict' and continuing 'train'
        for command in commands:
            if command == "prep":
                continue  # we don't run prep in this code block
            logger.info(f"running configs for command: {command}")
            command_config_metadata = [
                config_metadata
                for config_metadata in vaktestdata.constants.CONFIG_METADATA
                if config_metadata.config_type == command
            ]

            if command in ("predict", "eval", "train_continue"):
                # Fix values for required options in predict / eval / train_continue configs
                # using results from running the corresponding train configs.
                # this only works if we ran the train configs already,
                # which we should have because of ordering of COMMANDS constant above
                vaktestdata.configs.fix_options_in_configs(command_config_metadata, command, single_train_result)

            if command == "train_continue":
                # so we don't get 'command not recognized' error in next code block
                command = "train"

            for config_metadata in command_config_metadata:
                config_path = vaktestdata.constants.GENERATED_TEST_CONFIGS_ROOT / config_metadata.filename
                logger.info(
                    f"n\Running 'vak {command}', using config: {config_path.name}"
                )
                vak.cli.cli.cli(command, config_path)


if __name__ == "__main__":
    parser = vaktestdata.parser.get_parser()
    args = parser.parse_args()
    generate_test_data(
        step=args.step,
        commands=args.commands,
        single_train_result=args.single_train_result,
    )



from __future__ import annotations

import logging
import pathlib

from . import constants
from .frame_classification import prep_frame_classification_dataset
from .parametric_umap import prep_parametric_umap_dataset

logger = logging.getLogger(__name__)


def prep(
    data_dir: str | pathlib.Path,
    purpose: str,
    dataset_type: str,
    input_type: str,
    output_dir: str | pathlib.Path | None = None,
    audio_format: str | None = None,
    spect_format: str | None = None,
    spect_params: dict | None = None,
    annot_format: str | None = None,
    annot_file: str | pathlib.Path | None = None,
    labelset: set | None = None,
    audio_dask_bag_kwargs: dict | None = None,
    train_dur: int | None = None,
    val_dur: int | None = None,
    test_dur: int | None = None,
    train_set_durs: list[float] | None = None,
    num_replicates: int | None = None,
    spect_key: str = "s",
    timebins_key: str = "t",
    context_s: float = 0.015,
):
    """Prepare datasets for use with neural network models.

    Datasets are used to train and evaluate neural networks.
    The function also prepares datasets to generate predictions
    with trained neural networks, and to train a series of models
    with varying sizes of dataset so that performance
    can be evaluated as a function of dataset size,
    with a learning curve.

    When :func:`vak.core.prep` runs, it builds the dataset in
    ``data_dir`` containing the data itself as
    well as additional metadata.

    This is a high-level function that prepares datasets to be used by other
    high-level functions like :func:`vak.core.train`,
    :func:`vak.core.predict`, and :func:`vak.core.learncurve`.

    It can also split a dataset into training, validation, and test sets,
    e.g. for benchmarking different neural network architectures.
    If the ``purpose`` argument is set to 'train' or 'learncurve',
    and/or the duration of either the training or test set is provided,
    then the function attempts to split the dataset into training and test sets.
    A duration can also be specified for a validation set
    (used to measure performance during training).
    In these cases, the 'split' column in the .csv
    identifies which files (rows) belong to the training, test, and
    validation sets created from that Dataset.

    If the ``purpose`` is set to 'predict' or 'eval',
    or no durations for any of the training sets are specified,
    then the function assumes all the vocalizations constitute a single
    dataset, and for all rows the 'split' columns for that dataset
    will be 'predict' or 'test' (respectively).

    Parameters
    ----------
    data_dir : str, Path
        Path to directory with files from which to make dataset.
    purpose : str
        Purpose of the dataset.
        One of {'train', 'eval', 'predict', 'learncurve'}.
        These correspond to commands of the vak command-line interface.
    dataset_type : str
        String name of the type of dataset, e.g.,
        'frame_classification'. Dataset types are
        defined by machine learning tasks, e.g.,
        a 'frame_classification' dataset would be used
        a :class:`vak.models.FrameClassificationModel` model.
        Valid dataset types are defined as
        :const:`vak.prep.prep.DATASET_TYPES`.
    input_type : str
        The type of input to the neural network model.
        One of {'audio', 'spect'}.
    output_dir : str
        Path to location where data sets should be saved.
        Default is ``None``, in which case it defaults to ``data_dir``.
    audio_format : str
        Format of audio files. One of {'wav', 'cbin'}.
        Default is ``None``, but either ``audio_format`` or ``spect_format``
        must be specified.
    spect_format : str
        Format of files containing spectrograms as 2-d matrices. One of {'mat', 'npz'}.
        Default is None, but either audio_format or spect_format must be specified.
    spect_params : dict, vak.config.SpectParams
        Parameters for creating spectrograms. Default is ``None``.
    annot_format : str
        Format of annotations. Any format that can be used with the
        :mod:`crowsetta` library is valid. Default is ``None``.
    annot_file : str
        Path to a single annotation file. Default is ``None``.
        Used when a single file contains annotates multiple audio
        or spectrogram files.
    labelset : str, list, set
        Set of unique labels for vocalizations. Strings or integers.
        Default is ``None``. If not ``None``, then files will be skipped
        where the associated annotation
        contains labels not found in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using
        :func:`vak.converters.labelset_to_set`.
        See help for that function for details on how to specify ``labelset``.
    audio_dask_bag_kwargs : dict
        Keyword arguments used when calling :func:`dask.bag.from_sequence`
        inside :func:`vak.io.audio`, where it is used to parallelize
        the conversion of audio files into spectrograms.
        Option should be specified in config.toml file as an inline table,
        e.g., ``audio_dask_bag_kwargs = { npartitions = 20 }``.
        Allows for finer-grained control
        when needed to process files of different sizes.
    train_dur : float
        Total duration of training set, in seconds.
        When creating a learning curve,
        training subsets of shorter duration
        will be drawn from this set. Default is None.
    val_dur : float
        Total duration of validation set, in seconds.
        Default is None.
    test_dur : float
        Total duration of test set, in seconds.
        Default is None.
    train_set_durs : list
        of int, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5, 10, 15, 20].
    num_replicates : int
        number of times to replicate training for each training set duration
        to better estimate metrics for a training set of that size.
        Each replicate uses a different randomly drawn subset of the training
        data (but of the same duration).
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.

    Returns
    -------
    dataset_df : pandas.DataFrame
        That represents a dataset.
    dataset_path : pathlib.Path
        Path to csv saved from ``dataset_df``.
    """
    # pre-conditions ---------------------------------------------------------------------------------------------------
    if purpose not in constants.VALID_PURPOSES:
        raise ValueError(
            f"purpose must be one of: {constants.VALID_PURPOSES}\n"
            f"Value for purpose was: {purpose}"
        )

    if dataset_type not in constants.DATASET_TYPES:
        raise ValueError(
            f"``dataset_type`` must be one of: {constants.DATASET_TYPES}\n"
            f"Value for ``dataset_type`` was: {dataset_type}"
        )

    if input_type not in constants.INPUT_TYPES:
        raise ValueError(
            f"``input_type`` must be one of: {constants.INPUT_TYPES}\n"
            f"Value for ``input_type`` was: {input_type}"
        )

    if input_type == "audio" and spect_format is not None:
        raise ValueError(
            f"``input_type`` was set to 'audio' but a ``spect_format`` was specified: '{spect_format}'.\n"
            f"Please only provide a ``spect_format`` argument when the input type to the neural network "
            f"model is spectrograms."
        )

    if audio_format is None and spect_format is None:
        raise ValueError("Must specify either audio_format or spect_format")

    if audio_format and spect_format:
        raise ValueError(
            "Cannot specify both audio_format and spect_format, "
            "unclear whether to compute spectrograms from audio files or "
            "use pre-computed spectrograms from existing array files."
        )

    # we have to use an if-else here since args may vary across dataset prep functions
    # but we still define DATASET_TYPE_FUNC_MAP in vak.prep.constants
    # so that the mapping is made explicit in the code
    if dataset_type == "frame classification":
        dataset_df, dataset_path = prep_frame_classification_dataset(
            data_dir,
            input_type,
            purpose,
            output_dir,
            audio_format,
            spect_format,
            spect_params,
            annot_format,
            annot_file,
            labelset,
            audio_dask_bag_kwargs,
            train_dur,
            val_dur,
            test_dur,
            train_set_durs,
            num_replicates,
            spect_key,
            timebins_key,
        )
        return dataset_df, dataset_path
    elif dataset_type == "parametric umap":
        dataset_df, dataset_path = prep_parametric_umap_dataset(
            data_dir,
            purpose,
            output_dir,
            audio_format,
            spect_params,
            annot_format,
            annot_file,
            labelset,
            context_s,
            train_dur,
            val_dur,
            test_dur,
            train_set_durs,
            num_replicates,
            spect_key=spect_key,
            timebins_key=timebins_key,
        )
        return dataset_df, dataset_path
    else:
        # this is in case a dataset type is written wrong
        # in the if-else statements above, we want to error loudly
        raise ValueError(f"Unrecognized dataset type: {dataset_type}")



import pytest

# this constant defined "by hand" declares all the models for which there are fixtures,
# instead of computing something dynamically e.g. from ``vak.models``.
# Should be used throughout fixtures when we need to get things "by model"
MODELS = [
    "TweetyNet",
]


@pytest.fixture
def default_model():
    """default model used whenever a model is needed to run a test.
    Should work regardless of where the test is run, i.e. both on
    CI platform and locally.

    currently ``TweetyNet``
    """
    return "TweetyNet"



"""tests for vak.config.learncurve module"""
import pytest

import vak.config.learncurve


class TestLearncurveConfig:

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'standardize_frames': True,
                    'batch_size': 11,
                    'num_epochs': 2,
                    'val_step': 50,
                    'ckpt_step': 200,
                    'patience': 4,
                    'num_workers': 16,
                    'trainer': {'accelerator': 'gpu', 'devices': [0]},
                    'root_results_dir': './tests/data_for_tests/generated/results/train/audio_cbin_annot_notmat/TweetyNet',
                    'post_tfm_kwargs': {'majority_vote': True, 'min_segment_dur': 0.02},
                    'model': {
                        'TweetyNet': {
                            'network': {
                                'conv1_filters': 8,
                                'conv1_kernel_size': [3, 3],
                                'conv2_filters': 16,
                                'conv2_kernel_size': [5, 5],
                                'pool1_size': [4, 1],
                                'pool1_stride': [4, 1],
                                'pool2_size': [4, 1],
                                'pool2_stride': [4, 1],
                                'hidden_size': 32
                            },
                            'optimizer': {
                                'lr': 0.001
                            }
                        }
                    },
                    'dataset': {
                        'path': 'tests/data_for_tests/generated/prep/train/audio_cbin_annot_notmat/TweetyNet/032312-vak-frame-classification-dataset-generated-240502_234819'
                    }
                }
            ]
    )
    def test_init(self, config_dict):
        config_dict['model'] = vak.config.ModelConfig.from_config_dict(config_dict['model'])
        config_dict['dataset'] = vak.config.DatasetConfig.from_config_dict(config_dict['dataset'])
        config_dict['trainer'] = vak.config.TrainerConfig(**config_dict['trainer'])

        learncurve_config = vak.config.LearncurveConfig(**config_dict)

        assert isinstance(learncurve_config, vak.config.LearncurveConfig)

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'standardize_frames': True,
                    'batch_size': 11,
                    'num_epochs': 2,
                    'val_step': 50,
                    'ckpt_step': 200,
                    'patience': 4,
                    'num_workers': 16,
                    'trainer': {'accelerator': 'gpu', 'devices': [0]},
                    'root_results_dir': './tests/data_for_tests/generated/results/train/audio_cbin_annot_notmat/TweetyNet',
                    'post_tfm_kwargs': {'majority_vote': True, 'min_segment_dur': 0.02},
                    'model': {
                        'TweetyNet': {
                            'network': {
                                'conv1_filters': 8,
                                'conv1_kernel_size': [3, 3],
                                'conv2_filters': 16,
                                'conv2_kernel_size': [5, 5],
                                'pool1_size': [4, 1],
                                'pool1_stride': [4, 1],
                                'pool2_size': [4, 1],
                                'pool2_stride': [4, 1],
                                'hidden_size': 32
                            },
                            'optimizer': {
                                'lr': 0.001
                            }
                        }
                    },
                    'dataset': {
                        'path': 'tests/data_for_tests/generated/prep/train/audio_cbin_annot_notmat/TweetyNet/032312-vak-frame-classification-dataset-generated-240502_234819'
                    }
                }
            ]
    )
    def test_from_config_dict(self, config_dict):
        learncurve_config = vak.config.LearncurveConfig.from_config_dict(config_dict)

        assert isinstance(learncurve_config, vak.config.LearncurveConfig)

    def test_from_config_dict_with_real_config(self, a_generated_learncurve_config_dict):
        """test that instantiating LearncurveConfig class works as expected"""
        learncurve_table = a_generated_learncurve_config_dict["learncurve"]

        config = vak.config.learncurve.LearncurveConfig.from_config_dict(
            learncurve_table
        )

        assert isinstance(config, vak.config.learncurve.LearncurveConfig)

    @pytest.mark.parametrize(
            'config_dict, expected_exception',
            [
                # missing 'model', should raise KeyError
                (
                    {
                        'standardize_frames': True,
                        'batch_size': 11,
                        'num_epochs': 2,
                        'val_step': 50,
                        'ckpt_step': 200,
                        'patience': 4,
                        'num_workers': 16,
                        'trainer': {'accelerator': 'gpu', 'devices': [0]},
                        'root_results_dir': './tests/data_for_tests/generated/results/train/audio_cbin_annot_notmat/TweetyNet',
                        'post_tfm_kwargs': {'majority_vote': True, 'min_segment_dur': 0.02},
                        'dataset': {
                            'path': 'tests/data_for_tests/generated/prep/train/audio_cbin_annot_notmat/TweetyNet/032312-vak-frame-classification-dataset-generated-240502_234819'
                        }
                    },
                    KeyError
                ),
                # missing 'dataset', should raise KeyError
               (
                    {
                        'standardize_frames': True,
                        'batch_size': 11,
                        'num_epochs': 2,
                        'val_step': 50,
                        'ckpt_step': 200,
                        'patience': 4,
                        'num_workers': 16,
                        'trainer': {'accelerator': 'gpu', 'devices': [0]},
                        'root_results_dir': './tests/data_for_tests/generated/results/train/audio_cbin_annot_notmat/TweetyNet',
                        'post_tfm_kwargs': {'majority_vote': True, 'min_segment_dur': 0.02},
                        'model': {
                            'TweetyNet': {
                                'network': {
                                    'conv1_filters': 8,
                                    'conv1_kernel_size': [3, 3],
                                    'conv2_filters': 16,
                                    'conv2_kernel_size': [5, 5],
                                    'pool1_size': [4, 1],
                                    'pool1_stride': [4, 1],
                                    'pool2_size': [4, 1],
                                    'pool2_stride': [4, 1],
                                    'hidden_size': 32
                                },
                                'optimizer': {
                                    'lr': 0.001
                                }
                            }
                        },
                    },
                    KeyError
                ),
                # missing 'root_results_dir', should raise KeyError
                (
                    {
                        'standardize_frames': True,
                        'batch_size': 11,
                        'num_epochs': 2,
                        'val_step': 50,
                        'ckpt_step': 200,
                        'patience': 4,
                        'num_workers': 16,
                        'trainer': {'accelerator': 'gpu', 'devices': [0]},
                        'post_tfm_kwargs': {'majority_vote': True, 'min_segment_dur': 0.02},
                        'model': {
                            'TweetyNet': {
                                'network': {
                                    'conv1_filters': 8,
                                    'conv1_kernel_size': [3, 3],
                                    'conv2_filters': 16,
                                    'conv2_kernel_size': [5, 5],
                                    'pool1_size': [4, 1],
                                    'pool1_stride': [4, 1],
                                    'pool2_size': [4, 1],
                                    'pool2_stride': [4, 1],
                                    'hidden_size': 32
                                },
                                'optimizer': {
                                    'lr': 0.001
                                }
                            }
                        },
                        'dataset': {
                            'path': 'tests/data_for_tests/generated/prep/train/audio_cbin_annot_notmat/TweetyNet/032312-vak-frame-classification-dataset-generated-240502_234819'
                        }
                    },
                    KeyError
                )
            ]
    )
    def test_from_config_dict_raises(self, config_dict, expected_exception):
        with pytest.raises(expected_exception):
            vak.config.LearncurveConfig.from_config_dict(config_dict)



"""tests for vak.config.eval module"""
import pytest

import vak.config


class TestEval:

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'checkpoint_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/TweetyNet/checkpoints/max-val-acc-checkpoint.pt',
                    'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/labelmap.json',
                    'batch_size': 11,
                    'num_workers': 16,
                    'trainer': {'accelerator': 'gpu', 'devices': [0]},
                    'frames_standardizer_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/StandardizeSpect',
                    'output_dir': './tests/data_for_tests/generated/results/eval/audio_cbin_annot_notmat/TweetyNet',
                    'post_tfm_kwargs': {
                        'majority_vote': True, 'min_segment_dur': 0.02
                        },
                    'model': {
                        'TweetyNet': {
                            'network': {
                                'conv1_filters': 8,
                                'conv1_kernel_size': [3, 3],
                                'conv2_filters': 16,
                                'conv2_kernel_size': [5, 5],
                                'pool1_size': [4, 1],
                                'pool1_stride': [4, 1],
                                'pool2_size': [4, 1],
                                'pool2_stride': [4, 1],
                                'hidden_size': 32
                            },
                            'optimizer': {'lr': 0.001}
                            }
                        },
                    'dataset': {
                        'path': '~/some/path/I/made/up/for/now'
                    },
                }
            ]
    )
    def test_init(self, config_dict):
        config_dict['model'] = vak.config.ModelConfig.from_config_dict(config_dict['model'])
        config_dict['dataset'] = vak.config.DatasetConfig.from_config_dict(config_dict['dataset'])
        config_dict['trainer'] = vak.config.TrainerConfig(**config_dict['trainer'])

        eval_config = vak.config.EvalConfig(**config_dict)

        assert isinstance(eval_config, vak.config.EvalConfig)

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'checkpoint_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/TweetyNet/checkpoints/max-val-acc-checkpoint.pt',
                    'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/labelmap.json',
                    'batch_size': 11,
                    'num_workers': 16,
                    'trainer': {'accelerator': 'gpu', 'devices': [0]},
                    'frames_standardizer_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/StandardizeSpect',
                    'output_dir': './tests/data_for_tests/generated/results/eval/audio_cbin_annot_notmat/TweetyNet',
                    'post_tfm_kwargs': {
                        'majority_vote': True, 'min_segment_dur': 0.02
                        },
                    'model': {
                        'TweetyNet': {
                            'network': {
                                'conv1_filters': 8,
                                'conv1_kernel_size': [3, 3],
                                'conv2_filters': 16,
                                'conv2_kernel_size': [5, 5],
                                'pool1_size': [4, 1],
                                'pool1_stride': [4, 1],
                                'pool2_size': [4, 1],
                                'pool2_stride': [4, 1],
                                'hidden_size': 32
                            },
                            'optimizer': {'lr': 0.001}
                            }
                        },
                    'dataset': {
                        'path': '~/some/path/I/made/up/for/now'
                    },
                }
            ]
    )
    def test_from_config_dict(self, config_dict):
        eval_config = vak.config.EvalConfig.from_config_dict(config_dict)

        assert isinstance(eval_config, vak.config.EvalConfig)

    def test_from_config_dict_with_real_config(self, a_generated_eval_config_dict):
        eval_table = a_generated_eval_config_dict["eval"]

        eval_config = vak.config.eval.EvalConfig.from_config_dict(eval_table)

        assert isinstance(eval_config, vak.config.eval.EvalConfig)

    @pytest.mark.parametrize(
        'config_dict, expected_exception',
        [
            # missing 'model', should raise KeyError
            (
                {
                    'checkpoint_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/TweetyNet/checkpoints/max-val-acc-checkpoint.pt',
                    'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/labelmap.json',
                    'batch_size': 11,
                    'num_workers': 16,
                    'trainer': {'accelerator': 'gpu', 'devices': [0]},
                    'frames_standardizer_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/StandardizeSpect',
                    'output_dir': './tests/data_for_tests/generated/results/eval/audio_cbin_annot_notmat/TweetyNet',
                    'post_tfm_kwargs': {
                        'majority_vote': True, 'min_segment_dur': 0.02
                        },
                    'dataset': {
                        'path': '~/some/path/I/made/up/for/now'
                    },
                },
                KeyError
            ),
            # missing 'dataset', should raise KeyError
            (
                {
                    'checkpoint_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/TweetyNet/checkpoints/max-val-acc-checkpoint.pt',
                    'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/labelmap.json',
                    'batch_size': 11,
                    'num_workers': 16,
                    'trainer': {'accelerator': 'gpu', 'devices': [0]},
                    'frames_standardizer_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/StandardizeSpect',
                    'output_dir': './tests/data_for_tests/generated/results/eval/audio_cbin_annot_notmat/TweetyNet',
                    'post_tfm_kwargs': {
                        'majority_vote': True, 'min_segment_dur': 0.02
                        },
                    'model': {
                        'TweetyNet': {
                            'network': {
                                'conv1_filters': 8,
                                'conv1_kernel_size': [3, 3],
                                'conv2_filters': 16,
                                'conv2_kernel_size': [5, 5],
                                'pool1_size': [4, 1],
                                'pool1_stride': [4, 1],
                                'pool2_size': [4, 1],
                                'pool2_stride': [4, 1],
                                'hidden_size': 32
                            },
                            'optimizer': {'lr': 0.001}
                            }
                        },
                },
                KeyError
            ),
            # missing 'checkpoint_path', should raise KeyError
            (
                {
                    'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/labelmap.json',
                    'batch_size': 11,
                    'num_workers': 16,
                    'trainer': {'accelerator': 'gpu', 'devices': [0]},
                    'frames_standardizer_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/StandardizeSpect',
                    'output_dir': './tests/data_for_tests/generated/results/eval/audio_cbin_annot_notmat/TweetyNet',
                    'post_tfm_kwargs': {
                        'majority_vote': True, 'min_segment_dur': 0.02
                        },
                    'model': {
                        'TweetyNet': {
                            'network': {
                                'conv1_filters': 8,
                                'conv1_kernel_size': [3, 3],
                                'conv2_filters': 16,
                                'conv2_kernel_size': [5, 5],
                                'pool1_size': [4, 1],
                                'pool1_stride': [4, 1],
                                'pool2_size': [4, 1],
                                'pool2_stride': [4, 1],
                                'hidden_size': 32
                            },
                            'optimizer': {'lr': 0.001}
                            }
                        },
                    'dataset': {
                        'path': '~/some/path/I/made/up/for/now'
                    },
                },
                KeyError
            ),
            # missing 'output_dir', should raise KeyError
            (
                {
                    'checkpoint_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/TweetyNet/checkpoints/max-val-acc-checkpoint.pt',
                    'labelmap_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/labelmap.json',
                    'batch_size': 11,
                    'num_workers': 16,
                    'trainer': {'accelerator': 'gpu', 'devices': [0]},
                    'frames_standardizer_path': '~/Documents/repos/coding/birdsong/TweetyNet/results/BFSongRepository/gy6or6/results_200620_165308/StandardizeSpect',
                    'post_tfm_kwargs': {
                        'majority_vote': True, 'min_segment_dur': 0.02
                        },
                    'model': {
                        'TweetyNet': {
                            'network': {
                                'conv1_filters': 8,
                                'conv1_kernel_size': [3, 3],
                                'conv2_filters': 16,
                                'conv2_kernel_size': [5, 5],
                                'pool1_size': [4, 1],
                                'pool1_stride': [4, 1],
                                'pool2_size': [4, 1],
                                'pool2_stride': [4, 1],
                                'hidden_size': 32
                            },
                            'optimizer': {'lr': 0.001}
                            }
                        },
                    'dataset': {
                        'path': '~/some/path/I/made/up/for/now'
                    },
                },
                KeyError
            )
        ]
    )
    def test_from_config_dict_raises(self, config_dict, expected_exception):
        with pytest.raises(expected_exception):
            vak.config.EvalConfig.from_config_dict(config_dict)


"""Tests for vak.prep.audio_dataset module."""
from pathlib import Path

import pandas as pd
import pytest

from ..fixtures.audio import (
    AUDIO_DIR_CBIN, AUDIO_LIST_CBIN_ALL_LABELS_IN_LABELSET, AUDIO_LIST_CBIN_LABELS_NOT_IN_LABELSET
)
from ..fixtures.annot import LABELSET_NOTMAT

import vak.common.annotation
import vak.common.constants
import vak.prep.spectrogram_dataset.prep
import vak.prep.spectrogram_dataset.spect_helper


def returned_dataframe_matches_expected(
    df_returned,
    data_dir,
    labelset,
    audio_format,
    annot_format=None,
    annot_file=None,
    expected_audio_paths=None,
    not_expected_audio_paths=None,
):
    """tests that dataframe returned by ``vak.prep.audio_dataset.prep.prep_spectrogram_dataset``
    matches expected dataframe."""
    assert isinstance(df_returned, pd.DataFrame)

    assert df_returned.columns.values.tolist() == vak.prep.audio_dataset.DF_COLUMNS

    annot_format_from_df = df_returned.annot_format.unique()
    assert len(annot_format_from_df) == 1
    annot_format_from_df = annot_format_from_df[0]

    if annot_format is not None:
        # test format from dataframe is specified format
        assert annot_format_from_df == annot_format
        annot_list = vak.common.annotation.from_df(df_returned)
    else:
        # if no annotation format specified, check that `annot_format` is `none`
        assert annot_format_from_df == vak.common.constants.NO_ANNOTATION_FORMAT
        annot_list = None

    if annot_file:
        assert all(
            [Path(annot_path) == annot_file for annot_path in df_returned["annot_path"]]
        )

    if labelset:
        # test that filtering vocalizations by labelset worked
        assert annot_list is not None, "labelset specified but annot_list was None"
        for annot in annot_list:
            labels = annot.seq.labels
            labelset_from_labels = set(labels)
            # should be true that set of labels from each annotation is a subset of labelset
            assert labelset_from_labels.issubset(set(labelset))

    audio_paths_from_df = [Path(audio_path) for audio_path in df_returned.audio_path]

    # test that all audio files came from data_dir
    for audio_path_from_df in audio_paths_from_df:
        assert audio_path_from_df.parent == data_dir

    if expected_audio_paths:
        assert all(
            [
                expected_audio_path in audio_paths_from_df
                for expected_audio_path in expected_audio_paths
            ]
        )

    # if there are audio files we expect to **not** be in audio_path
    # -- because the associated annotations have labels not in labelset --
    # then test those files are **not** in spect_path
    if not_expected_audio_paths:
        assert all(
            [
                not_expected_audio_path not in audio_paths_from_df
                for not_expected_audio_path in not_expected_audio_paths
            ]
        )

    return True  # all asserts passed


@pytest.mark.parametrize(
    'data_dir, audio_format, labelset, annot_format, annot_file, expected_audio_paths, not_expected_audio_paths',
    [
        (AUDIO_DIR_CBIN, "cbin", LABELSET_NOTMAT, "notmat", None,
         AUDIO_LIST_CBIN_ALL_LABELS_IN_LABELSET, AUDIO_LIST_CBIN_LABELS_NOT_IN_LABELSET),
    ]
)
def test_prep_audio_dataset(
    data_dir, audio_format, labelset, annot_format, annot_file, tmp_path,
    expected_audio_paths, not_expected_audio_paths,
):
    """test that ``vak.prep.spectrogram_dataset.prep.prep_spectrogram_dataset`` works
    when we point it at directory of .cbin audio files
    and specify an annotation format"""
    dataset_df = vak.prep.audio_dataset.prep_audio_dataset(
        data_dir=data_dir,
        audio_format=audio_format,
        labelset=labelset,
        annot_format=annot_format,
        annot_file=annot_file,
    )

    assert returned_dataframe_matches_expected(
        dataset_df,
        data_dir=data_dir,
        labelset=labelset,
        annot_format=annot_format,
        audio_format=audio_format,
        annot_file=annot_file,
        expected_audio_paths=expected_audio_paths,
        not_expected_audio_paths=not_expected_audio_paths,
    )



"""Unit tests for vak.prep.frame_classification.make_splits"""
import json
import pathlib
import shutil

import crowsetta
import numpy as np
import pandas as pd
import pytest

import vak.prep.frame_classification.make_splits


@pytest.mark.parametrize(
    'annots, expected_sort_inds',
    [
        (
            [
                crowsetta.Annotation(seq=crowsetta.Sequence.from_keyword(
                    onsets_s=[0.1, 0.3, 0.5], offsets_s=[0.2, 0.4, 0.6], labels=['a', 'b', 'b']
                ), annot_path='./fake'),
                crowsetta.Annotation(seq=crowsetta.Sequence.from_keyword(
                    onsets_s=[0.1, 0.3, 0.5], offsets_s=[0.2, 0.4, 0.6], labels=['b', 'b', 'b']
                ), annot_path='./fake'),
            ],
            [0, 1,]
        ),
        (
            [
                crowsetta.Annotation(seq=crowsetta.Sequence.from_keyword(
                    onsets_s=[0.1, 0.3, 0.5], offsets_s=[0.2, 0.4, 0.6], labels=['a', 'b', 'b']
                ), annot_path='./fake'),
                crowsetta.Annotation(seq=crowsetta.Sequence.from_keyword(
                    onsets_s=[0.1, 0.3, 0.5], offsets_s=[0.2, 0.4, 0.6], labels=['b', 'b', 'b']
                ), annot_path='./fake'),
                crowsetta.Annotation(seq=crowsetta.Sequence.from_keyword(
                    onsets_s=[0.1, 0.3, 0.5], offsets_s=[0.2, 0.4, 0.6], labels=['b', 'b', 'b']
                ), annot_path='./fake'),
            ],
            [0, 1, 2],
        ),
    ]
)
def test_argsort_by_label_freq(annots, expected_sort_inds):
    out = vak.prep.frame_classification.make_splits.argsort_by_label_freq(annots)
    assert isinstance(out, list)
    assert out == expected_sort_inds


def copy_dataset_df_files_to_tmp_path_data_dir(dataset_df, dataset_path, config_type, input_type, tmp_path_data_dir):
    """Copy all the files in a dataset DataFrame to a `tmp_path_data_dir`,
    and change the paths in the Dataframe, so that we can then call
    `vak.prep.frame_classification.helper.move_files_into_split_subdirs`."""
    paths_cols = []
    if input_type == 'spect':
        paths_cols.append('spect_path')
    elif input_type == 'audio':
        paths_cols.append('audio_path')
    if config_type != 'predict':
        paths_cols.append('annot_path')
    for paths_col in paths_cols:
        paths = dataset_df[paths_col].values
        new_paths = []
        for path in paths:
            new_path = shutil.copy(src=dataset_path / path, dst=tmp_path_data_dir)
            new_paths.append(new_path)
        dataset_df[paths_col] = new_paths
    return dataset_df


@pytest.mark.parametrize(
    'config_type, model_name, audio_format, spect_format, annot_format, input_type',
    [
        ('train', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        ('predict', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        ('eval', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        ('train', 'TweetyNet', None, 'mat', 'yarden', 'spect'),
        ('learncurve', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        # TODO: add audio cases
    ]
)
def test_make_splits(config_type, model_name, audio_format, spect_format, annot_format,
                     input_type, tmp_path, specific_config_toml_path, specific_source_files_with_split_df):
    toml_path = specific_config_toml_path(
        config_type,
        model_name,
        annot_format,
        audio_format,
        spect_format,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    # ---- set up ----
    tmp_dataset_path = tmp_path / 'dataset_dir'
    tmp_dataset_path.mkdir()

    purpose = config_type

    dataset_df = specific_source_files_with_split_df(
        config_type,
        model_name,
        annot_format,
        audio_format,
        spect_format,
    )
    if purpose != "predict":
        # TODO: add option to generate predict using existing dataset, so we can get labelmap from it
        map_unlabeled_segments = vak.prep.sequence_dataset.has_unlabeled_segments(
            dataset_df
        )
        labelmap = vak.common.labels.to_map(
            cfg.prep.labelset, map_background=map_unlabeled_segments
        )
    else:
        labelmap = None

    dataset_df_with_splits = vak.prep.frame_classification.make_splits.make_splits(
        dataset_df,
        tmp_dataset_path,
        cfg.prep.input_type,
        purpose,
        labelmap,
        cfg.prep.audio_format,
    )
    assert isinstance(dataset_df_with_splits, pd.DataFrame)

    splits = [
        split
        for split in sorted(dataset_df_with_splits.split.dropna().unique())
        if split != "None"
    ]

    for split in splits:
        split_subdir = tmp_dataset_path / split
        if split != 'None':
            assert split_subdir.exists()
        elif split == 'None':
            assert not split_subdir.exists()

        split_df = dataset_df_with_splits[
            dataset_df_with_splits.split == split
        ].copy()

        assert vak.datapipes.frame_classification.constants.FRAMES_PATH_COL_NAME in split_df.columns

        frames_paths = split_df[
            vak.datapipes.frame_classification.constants.FRAMES_PATH_COL_NAME
        ].values

        if purpose != "predict":
            assert vak.datapipes.frame_classification.constants.MULTI_FRAME_LABELS_PATH_COL_NAME in split_df.columns

            frame_labels_paths = split_df[
                vak.datapipes.frame_classification.constants.MULTI_FRAME_LABELS_PATH_COL_NAME
            ].values

            annots = vak.common.annotation.from_df(split_df)

            frames_tuples = [
                (frames_path, frame_labels_path, annot)
                for frames_path, frame_labels_path, annot in zip(
                    frames_paths, frame_labels_paths, annots
                )
            ]
        else:
            frames_tuples = [
                (frames_path, None, None)
                for frames_path in frames_paths
            ]

        sample_id_vecs, inds_in_sample_vecs = [], []
        for sample_id, frames_tuple in enumerate(frames_tuples):
            frames_path, frame_labels_path, annot = frames_tuple
            frames_file_that_should_exist = tmp_dataset_path / frames_path
            assert frames_file_that_should_exist.exists()

            # NOTE we load frames to confirm we can and also to make indexing vectors we use to test,
            # see next code block
            frames = vak.datapipes.frame_classification.helper.load_frames(tmp_dataset_path / frames_path, input_type)
            assert isinstance(frames, np.ndarray)

            # make indexing vectors that we use to test
            n_frames = frames.shape[-1]
            sample_id_vecs.append(np.ones((n_frames,)).astype(np.int32) * sample_id)
            inds_in_sample_vecs.append(np.arange(n_frames))

            if frame_labels_path is not None and annot is not None:
                frame_labels_file_that_should_exist = tmp_dataset_path / frame_labels_path
                assert frame_labels_file_that_should_exist.exists()

                if input_type == "audio":
                    _, samplefreq = vak.common.constants.AUDIO_FORMAT_FUNC_MAP[
                        audio_format
                    ](tmp_dataset_path / frames_path)
                    frame_times = np.arange(frames.shape[-1]) / samplefreq
                elif input_type == "spect":
                    spect_dict = vak.common.files.spect.load(tmp_dataset_path / frames_path, "npz")
                    frame_times = spect_dict[vak.common.constants.TIMEBINS_KEY]

                lbls_int = [labelmap[lbl] for lbl in annot.seq.labels]
                expected_frame_labels = vak.transforms.frame_labels.from_segments(
                    lbls_int,
                    annot.seq.onsets_s,
                    annot.seq.offsets_s,
                    frame_times,
                    background_label=labelmap[vak.common.constants.DEFAULT_BACKGROUND_LABEL],
                )
                frame_labels = np.load(frame_labels_file_that_should_exist)
                assert np.array_equal(frame_labels, expected_frame_labels)

        # assert there are no remaining .spect.npz files in dataset path (root)
        # because they were moved in to splits, and we removed any remaining that were not put into splits
        spect_npz_files_not_in_split = sorted(
            tmp_dataset_path.glob(f'*{vak.common.constants.SPECT_NPZ_EXTENSION}')
        )
        assert len(spect_npz_files_not_in_split) == 0

        sample_id_vec_path = (
            split_subdir /
            vak.datapipes.frame_classification.constants.SAMPLE_IDS_ARRAY_FILENAME
        )
        assert sample_id_vec_path.exists()

        expected_sample_id_vec = np.concatenate(sample_id_vecs)
        sample_id_vec = np.load(sample_id_vec_path)
        assert np.array_equal(sample_id_vec, expected_sample_id_vec)

        inds_in_sample_vec_path = (
            split_subdir /
            vak.datapipes.frame_classification.constants.INDS_IN_SAMPLE_ARRAY_FILENAME
        )
        assert inds_in_sample_vec_path.exists()

        expected_inds_in_sample_vec = np.concatenate(inds_in_sample_vecs)
        inds_in_sample_vec = np.load(inds_in_sample_vec_path)
        assert np.array_equal(inds_in_sample_vec, expected_inds_in_sample_vec)



import pytest
import torch


TRAINER_TABLE = [
    {"accelerator": "cpu", "devices": 1}
]
if torch.cuda.is_available():
    {"accelerator": "gpu", "devices": [0]}


@pytest.fixture(params=TRAINER_TABLE)
def trainer_table(request):
    """Parametrized 'trainer' table for config file

    Causes any test using this :func:`trainer_table` fixture
    to run just once if only a cpu is available,
    and twice if ``torch.cuda.is_available()`` returns ``True``."""
    return request.param



import pandas as pd
import pytest

import vak


@pytest.mark.parametrize(
    'model_name, config_type, audio_format, spect_format, annot_format, expected_result',
    [
        ("TweetyNet", "train", "cbin", None, "notmat", True),
        ("TweetyNet", "train", None, "mat", "yarden", True),
    ]
)
def test_has_unlabeled_segments(config_type,
                                model_name,
                                audio_format,
                                spect_format,
                                annot_format,
                                expected_result,
                                specific_config_toml,
                                specific_dataset_csv_path):
    dataset_csv_path = specific_dataset_csv_path(config_type,
                                                 model_name,
                                                 annot_format,
                                                 audio_format,
                                                 spect_format)

    dataset_df = pd.read_csv(dataset_csv_path)
    has_unlabeled = vak.prep.sequence_dataset.has_unlabeled_segments(dataset_df)
    assert has_unlabeled == expected_result



"""Module that contains datapipe classes,
used to load inputs and targets for neural network models
from datasets prepared by :func:`vak.prep.prep`"""

from . import frame_classification, parametric_umap

__all__ = ["frame_classification", "parametric_umap"]



FRAMES_PATH_COL_NAME = "frames_path"
MULTI_FRAME_LABELS_EXT = ".multi-frame-labels.npy"
MULTI_FRAME_LABELS_PATH_COL_NAME = "multi_frame_labels_path"
ANNOTATION_CSV_FILENAME = "y.csv"
SAMPLE_IDS_ARRAY_FILENAME = "sample_ids.npy"
INDS_IN_SAMPLE_ARRAY_FILENAME = "inds_in_sample.npy"
WINDOW_INDS_ARRAY_FILENAME = "window_inds.npy"
FRAME_CLASSIFICATION_DATASET_AUDIO_FORMAT = "wav"
